# -----------------------------------------------------------------------------
# Copyright (c) 2021- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Spyder API configuration helper mixins.
"""

from __future__ import annotations

# Standard library imports
import logging
import sys
import warnings
from collections.abc import Callable
from typing import Union

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias  # noqa: ICN003

# Third-party imports
from qtpy import PYSIDE6

# Local imports
from spyder.config.manager import CONF, CONF_VERSION
from spyder.config.types import ConfigurationKey
from spyder.config.user import NoDefault


logger = logging.getLogger(__name__)

BasicTypes: TypeAlias = Union[bool, int, str, tuple, list, dict]
"""Type alias for the set of basic Python types supported as config values."""


class SpyderConfigurationAccessor:
    """Mixin to access options stored in the Spyder configuration system."""

    CONF_SECTION: str | None = None
    """Name of the default configuration section to use for this object.

    Will be used to record its permanent data in Spyder's config system.
    """

    def get_conf(
        self,
        option: ConfigurationKey,
        default: NoDefault | BasicTypes = NoDefault,
        section: str | None = None,
        secure: bool = False,
    ) -> BasicTypes:
        """
        Retrieve an option's value from the Spyder configuration system.

        Parameters
        ----------
        option: spyder.config.types.ConfigurationKey
            Name/tuple path of the configuration option value to get.
        default: spyder.config.user.NoDefault | BasicTypes, optional
            Fallback value to return if the option is not found on the
            configuration system. No default value if not passed.
        section: str | None, optional
            Name of the configuration section to use, e.g. ``"shortcuts"``.
            If ``None``, then the value of :attr:`CONF_SECTION` is used.
        secure: bool, optional
            If ``True``, the option will be retrieved from secure storage
            using the :mod:`!keyring` Python package. Otherwise, will be
            retrieved from Spyder's normal configuration (the default).

        Returns
        -------
        value: BasicTypes
            Value of ``option`` in the configuration ``section``.

        Raises
        ------
        configparser.NoOptionError
            If the ``section`` does not exist in Spyder's configuration.
        """
        section = self.CONF_SECTION if section is None else section
        if section is None:
            raise AttributeError(
                "A SpyderConfigurationAccessor must define a `CONF_SECTION` "
                "class attribute!"
            )

        return CONF.get(section, option, default, secure)

    def get_conf_options(self, section: str | None = None) -> list[str]:
        """
        Get all option names from the given section.

        Parameters
        ----------
        section: str | None, optional
            Name of the configuration section to use, e.g. ``"shortcuts"``.
            If ``None``, then the value of :attr:`CONF_SECTION` is used.

        Returns
        -------
        values: list[str]
            List of option names (keys) in the configuration ``section``.

        Raises
        ------
        configparser.NoOptionError
            If ``section`` does not exist in the configuration.
        """
        section = self.CONF_SECTION if section is None else section
        if section is None:
            raise AttributeError(
                "A SpyderConfigurationAccessor must define a `CONF_SECTION` "
                "class attribute!"
            )
        return CONF.options(section)

    def set_conf(
        self,
        option: ConfigurationKey,
        value: BasicTypes,
        section: str | None = None,
        recursive_notification: bool = True,
        secure: bool = False,
    ) -> None:
        """
        Set an option's value in the Spyder configuration system.

        Parameters
        ----------
        option: spyder.config.types.ConfigurationKey
            Name/tuple path of the configuration option to set.
        value: BasicTypes
            Value to set for the given configuration option.
        section: str | None, optional
            Name of the configuration section to use, e.g. ``"shortcuts"``.
            If ``None``, then the value of :attr:`CONF_SECTION` is used.
        recursive_notification: bool, optional
            If ``True``, all objects that observe all changes on the
            configuration ``section`` as well as objects that observe
            partial tuple paths are notified. For example, if the
            ``option`` ``"opt"`` of ``section`` ``"sec"`` changes, then
            all observers for section ``sec`` are notified. Likewise,
            if the option ``("a", "b", "c")`` changes, then observers for
            ``("a", "b", "c")``, ``("a", "b")`` and ``"a"`` are all notified.
        secure: bool, optional
            If ``True``, the option will be saved in secure storage
            using the :mod:`!keyring` Python package. Otherwise, will be
            saved in Spyder's normal configuration (the default).

        Returns
        -------
        None
        """
        section = self.CONF_SECTION if section is None else section
        if section is None:
            raise AttributeError(
                "A SpyderConfigurationAccessor must define a `CONF_SECTION` "
                "class attribute!"
            )
        CONF.set(
            section,
            option,
            value,
            recursive_notification=recursive_notification,
            secure=secure,
        )

    def remove_conf(
        self,
        option: ConfigurationKey,
        section: str | None = None,
        secure: bool = False,
    ) -> None:
        """
        Remove an option from the Spyder configuration system.

        Parameters
        ----------
        option: spyder.config.types.ConfigurationKey
            Name/tuple path of the configuration option to remove.
        section: str | None, optional
            Name of the configuration section to use, e.g. ``"shortcuts"``.
            If ``None``, then the value of :attr:`CONF_SECTION` is used.
        secure: bool, optional
            If ``True``, the option will be removed from secure storage
            using the :mod:`!keyring` Python package. Otherwise, will be
            removed from Spyder's normal configuration (the default).

        Returns
        -------
        None
        """
        section = self.CONF_SECTION if section is None else section
        if section is None:
            raise AttributeError(
                "A SpyderConfigurationAccessor must define a `CONF_SECTION` "
                "class attribute!"
            )
        CONF.remove_option(section, option, secure)

    def get_conf_default(
        self,
        option: ConfigurationKey,
        section: str | None = None,
    ) -> Union[NoDefault, BasicTypes]:
        """
        Get an option's default value from the Spyder configuration system.

        Parameters
        ----------
        option: spyder.config.types.ConfigurationKey
            Name/tuple path of the config option to get the default value of.
        section: str | None, optional
            Name of the configuration section to use, e.g. ``"shortcuts"``.
            If ``None``, then the value of :attr:`CONF_SECTION` is used.

        Returns
        -------
        spyder.config.user.NoDefault | BasicTypes
            The ``option``'s default value, or
            :class:`spyder.config.user.NoDefault` if one is not set.
        """
        section = self.CONF_SECTION if section is None else section
        if section is None:
            raise AttributeError(
                "A SpyderConfigurationAccessor must define a `CONF_SECTION` "
                "class attribute!"
            )
        return CONF.get_default(section, option)

    @property
    def spyder_conf_version(self) -> str:
        """Get current version of the Spyder configuration system.

        Returns
        -------
        str
            The current Spyder :const:`spyder.config.manager.CONF_VERSION`,
            as a string of ``MAJOR.MINOR.MICRO``.
        """
        return CONF_VERSION

    @property
    def old_spyder_conf_version(self) -> str:
        """Get previous version of the Spyder configuration system.

        Returns
        -------
        str
            The previous Spyder :const:`spyder.config.manager.CONF_VERSION`
            prior to the most recent Spyder update, as a string of
            ``MAJOR.MINOR.MICRO``.
        """
        return CONF.old_spyder_version


class SpyderConfigurationObserver(SpyderConfigurationAccessor):
    """
    Methods to receive and respond to changes in Spyder's configuration.

    This mixin enables a class to receive configuration updates seamlessly,
    by registering methods using the
    :func:`~spyder.api.config.decorators.on_conf_change` decorator,
    which takes a configuration section and option to observe.

    When a change occurs on any of the registered configuration options,
    the corresponding registered method is called with the new config value.
    """

    def __init__(self) -> None:
        """Create a new :class:`!SpyderConfigurationObserver`.

        .. important::

            Classes or their parents implementing this mixin must define
            a :attr:`~SpyderConfigurationAccessor.CONF_SECTION` class attribute
            with the name of their default configuration section.

        Returns
        -------
        None
        """
        super().__init__()
        if self.CONF_SECTION is None:
            warnings.warn(
                "A SpyderConfigurationObserver must define a `CONF_SECTION` "
                f"class attribute! Hint: {self} or its parent should define "
                "the section."
            )

        self._configuration_listeners = {}
        self._multi_option_listeners = set({})
        self._gather_observers()
        self._merge_none_observers()

        # Register class to listen for changes in all registered options
        for section in self._configuration_listeners:
            section = self.CONF_SECTION if section is None else section
            observed_options = self._configuration_listeners[section]
            for option in observed_options:
                # Avoid a crash at startup due to MRO
                if not PYSIDE6:
                    logger.debug(
                        f'{self} is observing option "{option}" in section '
                        f'"{section}"'
                    )

                CONF.observe_configuration(self, section, option)

    def __del__(self) -> None:
        """Remove an object from the configuration observer."""
        CONF.unobserve_configuration(self)

    def _gather_observers(self):
        """Gather all the methods decorated with :func:`on_conf_change`."""
        for method_name in dir(self):
            # Avoid crash at startup due to MRO
            if PYSIDE6 and method_name in {
                # PySide seems to require that the class is instantiated to
                # access this method
                "painters",
                # Method is debounced
                "restart_kernel",
            }:
                continue

            method = getattr(self, method_name, None)
            if hasattr(method, "_conf_listen"):
                info = method._conf_listen
                if len(info) > 1:
                    self._multi_option_listeners |= {method_name}

                for section, option in info:
                    self._add_listener(method_name, option, section)

    def _merge_none_observers(self):
        """Replace section ``None`` with ``CONF_SECTION`` in observers."""
        default_selectors = self._configuration_listeners.get(None, {})
        section_selectors = self._configuration_listeners.get(
            self.CONF_SECTION, {}
        )

        for option in default_selectors:
            default_option_receivers = default_selectors.get(option, [])
            section_option_receivers = section_selectors.get(option, [])
            merged_receivers = (
                default_option_receivers + section_option_receivers
            )
            section_selectors[option] = merged_receivers

        self._configuration_listeners[self.CONF_SECTION] = section_selectors
        self._configuration_listeners.pop(None, None)

    def _add_listener(
        self, func: Callable, option: ConfigurationKey, section: str
    ):
        """
        Add a callable as a listener of a specific configuration option.

        Parameters
        ----------
        func: Callable
            Function/method that will be called when ``option`` changes.
        option: spyder.config.types.ConfigurationKey
            Name/tuple path of the configuration option to observe.
        section: str
            Name of the section containing ``option``, e.g. ``"shortcuts"``.

        Returns
        -------
        None
        """
        section_listeners = self._configuration_listeners.get(section, {})
        option_listeners = section_listeners.get(option, [])
        option_listeners.append(func)
        section_listeners[option] = option_listeners
        self._configuration_listeners[section] = section_listeners

    def on_configuration_change(
        self, option: ConfigurationKey, section: str, value: BasicTypes
    ) -> None:
        """
        Handle configuration value updates for a config option.

        Parameters
        ----------
        option: spyder.config.types.ConfigurationKey
            Name/tuple path of the configuration option that changed.
        section: str
            Name of the section containing ``option``, e.g. ``"shortcuts"``.
        value: BasicTypes
            New value of the configuration option that produced the event.

        Returns
        -------
        None
        """
        section_receivers = self._configuration_listeners.get(section, {})
        option_receivers = section_receivers.get(option, [])
        for receiver in option_receivers:
            method = (
                receiver if callable(receiver) else getattr(self, receiver)
            )
            if receiver in self._multi_option_listeners:
                method(option, value)
            else:
                method(value)

    def add_configuration_observer(
        self,
        func: Callable,
        option: ConfigurationKey,
        section: str | None = None,
    ) -> None:
        """
        Add a callable to observe changes to a specific configuration option.

        Parameters
        ----------
        func: Callable
            Function/method that will be called when ``option`` changes.
        option: spyder.config.types.ConfigurationKey
            Name/tuple path of the configuration option to observe.
        section: str | None, optional
            Name of the section containing ``option``, e.g. ``"shortcuts"``.
            If ``None``, then the value of
            :attr:`~SpyderConfigurationAccessor.CONF_SECTION` is used.

        Returns
        -------
        None

        Notes
        -----
        - This is only necessary if you need to add a callable that is not a
          class method to observe an option. Otherwise, you only need to
          decorate your method with
          :func:`~spyder.api.config.decorators.on_conf_change`.
        """
        if section is None:
            section = self.CONF_SECTION

        logger.debug(
            f'{self} is observing "{option}" option on section "{section}"'
        )
        self._add_listener(func, option, section)
        CONF.observe_configuration(self, section, option)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Language Server Protocol linting configuration tab.
"""

# Standard library imports
import re

# Third party imports
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QButtonGroup,
    QGroupBox,
    QGridLayout,
    QLabel,
    QMessageBox,
    QVBoxLayout,
    QWidget,
)

# Local imports
from spyder.api.preferences import SpyderPreferencesTab
from spyder.api.translations import _
from spyder.utils.stylesheet import AppStyle


class LintingConfigTab(SpyderPreferencesTab):
    """Linting configuration tab."""

    TITLE = _('Linting')

    def __init__(self, parent):
        super().__init__(parent)
        newcb = self.create_checkbox

        linting_label = QLabel(
            _(
                "Spyder can highlight syntax errors and possible problems "
                "with your code in the editor by using one of the providers "
                "below"
            )
        )
        linting_label.setOpenExternalLinks(True)
        linting_label.setWordWrap(True)

        linting_select_group = QGroupBox(_("Provider"))

        linting_bg = QButtonGroup(linting_select_group)
        basic_linting_radio = self.create_radiobutton(
            _("Pyflakes (Basic)"),
            'pyflakes',
            button_group=linting_bg
        )
        flake_linting_radio = self.create_radiobutton(
            _("Flake8 (Intermediate)"),
            'flake8',
            button_group=linting_bg
        )
        ruff_linting_radio = self.create_radiobutton(
            _("Ruff (Advanced)"),
            'ruff',
            button_group=linting_bg
        )
        disable_linting_radio = self.create_radiobutton(
            _("Disable linting"),
            'no_linting',
            button_group=linting_bg
        )

        linting_select_layout = QVBoxLayout()
        linting_select_layout.addSpacing(3 * AppStyle.MarginSize)
        linting_select_layout.addWidget(basic_linting_radio)
        linting_select_layout.addWidget(flake_linting_radio)
        linting_select_layout.addWidget(ruff_linting_radio)
        linting_select_layout.addWidget(disable_linting_radio)
        linting_select_group.setLayout(linting_select_layout)

        additional_options_group = QGroupBox(_("Additional options"))

        underline_errors_box = newcb(
            _("Underline errors and warnings"),
            'underline_errors',
            section='editor')

        additional_options_layout = QVBoxLayout()
        additional_options_layout.addWidget(underline_errors_box)
        additional_options_group.setLayout(additional_options_layout)

        configuration_options_group = QGroupBox(_("Provider options"))
        configuration_options_layout = QVBoxLayout()

        # ruff options
        # This needs to be an attribute of the config page since it is used
        # for `test_mainwindow.py::test_preferences_checkboxes_not_checked_regression`
        self.docstring_style_check = self.create_checkbox(
            _("Enable docstring style linting"), "pydocstyle"
        )
        docstring_style_convention = self.create_combobox(
            _("Convention used to lint docstrings: "),
            (
                ("Numpy", "numpy"),
                ("PEP 257", "pep257"),
                ("Google", "google"),
            ),
            "pydocstyle/convention",
        )
        docstring_style_convention.label.setEnabled(
            self.get_option('pydocstyle')
        )
        docstring_style_convention.combobox.setEnabled(
            self.get_option('pydocstyle')
        )
        self.docstring_style_check.checkbox.toggled.connect(
            docstring_style_convention.label.setEnabled
        )
        self.docstring_style_check.checkbox.toggled.connect(
            docstring_style_convention.combobox.setEnabled
        )

        self.ruff_exclude = self.create_lineedit(
            _("Exclude these files or directories:"),
            'ruff/exclude',
            alignment=Qt.Horizontal,
            word_wrap=False,
            placeholder=_("Exclude test files: (?!test_).*\\.py"),
        )

        ruff_select = self.create_lineedit(
            _("Show these errors or warnings:"),
            'ruff/extendSelect',
            alignment=Qt.Horizontal,
            word_wrap=False,
            placeholder=_("Example codes: E113, W391"),
        )

        ruff_ignore = self.create_lineedit(
            _("Ignore these errors or warnings:"),
            'ruff/extendIgnore',
            alignment=Qt.Horizontal,
            word_wrap=False,
            placeholder=_("Default is: E"),
        )

        ruff_layout = QGridLayout()
        ruff_layout.addWidget(self.docstring_style_check)
        ruff_layout.addWidget(docstring_style_convention.label, 1, 0)
        ruff_layout.addWidget(docstring_style_convention.combobox, 1, 1)
        ruff_layout.addWidget(self.ruff_exclude.label, 2, 0)
        ruff_layout.addWidget(self.ruff_exclude.textbox, 2, 1)
        ruff_layout.addWidget(ruff_select.label, 3, 0)
        ruff_layout.addWidget(ruff_select.textbox, 3, 1)
        ruff_layout.addWidget(ruff_ignore.label, 4, 0)
        ruff_layout.addWidget(ruff_ignore.textbox, 4, 1)

        ruff_grid_widget = QWidget()
        ruff_grid_widget.setLayout(ruff_layout)

        # Flake8 options
        self.flake8_filenames_match = self.create_lineedit(
            _("Only check these filenames:"),
            'flake8/filename',
            alignment=Qt.Horizontal,
            word_wrap=False,
            placeholder=_("Check test files: test_.*\\.py"),
        )

        self.flake8_exclude = self.create_lineedit(
            _("Exclude these files or directories:"),
            'flake8/exclude',
            alignment=Qt.Horizontal,
            word_wrap=False,
            placeholder=_("Exclude test files: (?!test_).*\\.py"),
        )

        flake8_select = self.create_lineedit(
            _("Show these errors or warnings:"),
            'flake8/extendSelect',
            alignment=Qt.Horizontal,
            word_wrap=False,
            placeholder=_("Example codes: E113, W391"),
        )

        flake8_ignore = self.create_lineedit(
            _("Ignore these errors or warnings:"),
            'flake8/extendIgnore',
            alignment=Qt.Horizontal,
            word_wrap=False,
            placeholder=_("Default is: E,W,C90"),
        )

        flake8_layout = QGridLayout()
        flake8_layout.addWidget(self.flake8_filenames_match.label, 1, 0)
        flake8_layout.addWidget(self.flake8_filenames_match.textbox, 1, 1)
        flake8_layout.addWidget(self.flake8_exclude.label, 2, 0)
        flake8_layout.addWidget(self.flake8_exclude.textbox, 2, 1)
        flake8_layout.addWidget(flake8_select.label, 3, 0)
        flake8_layout.addWidget(flake8_select.textbox, 3, 1)
        flake8_layout.addWidget(flake8_ignore.label, 4, 0)
        flake8_layout.addWidget(flake8_ignore.textbox, 4, 1)
        flake8_grid_widget = QWidget()
        flake8_grid_widget.setLayout(flake8_layout)

        # pyflakes options
        pyflakes_conf_options = QLabel(
            _("There are no configuration options for Pyflakes")
        )

        # Disabled linting options
        not_select_conf_options = QLabel(_("Linting is disabled"))

        configuration_options_layout.addWidget(ruff_grid_widget)
        configuration_options_layout.addWidget(flake8_grid_widget)
        configuration_options_layout.addWidget(pyflakes_conf_options)
        configuration_options_layout.addWidget(not_select_conf_options)

        ruff_linting_radio.radiobutton.toggled.connect(
            lambda checked: (
                ruff_grid_widget.setVisible(checked),
                flake8_grid_widget.setVisible(False),
                pyflakes_conf_options.setVisible(False),
                not_select_conf_options.setVisible(False)
            ) if checked else None
        )

        flake_linting_radio.radiobutton.toggled.connect(
            lambda checked: (
                ruff_grid_widget.setVisible(False),
                flake8_grid_widget.setVisible(checked),
                pyflakes_conf_options.setVisible(False),
                not_select_conf_options.setVisible(False)
            ) if checked else None
        )

        basic_linting_radio.radiobutton.toggled.connect(
            lambda checked: (
                ruff_grid_widget.setVisible(False),
                flake8_grid_widget.setVisible(False),
                pyflakes_conf_options.setVisible(checked),
                not_select_conf_options.setVisible(False)
            ) if checked else None
        )

        disable_linting_radio.radiobutton.toggled.connect(
            lambda checked: (
                ruff_grid_widget.setVisible(False),
                flake8_grid_widget.setVisible(False),
                pyflakes_conf_options.setVisible(False),
                not_select_conf_options.setVisible(checked)
            ) if checked else None
        )

        configuration_options_group.setLayout(configuration_options_layout)

        # Linting layout
        linting_layout = QVBoxLayout()
        linting_layout.addWidget(linting_label)
        linting_layout.addWidget(linting_select_group)
        linting_layout.addWidget(configuration_options_group)
        linting_layout.addWidget(additional_options_group)
        self.setLayout(linting_layout)

    def report_invalid_regex(self, files=True):
        """
        Report that excluded files/directories should be valid regular
        expressions.
        """
        msg = _(
            "Directory patterns listed for exclusion should be valid regular "
            "expressions"
        )
        if files:
            msg = _(
                "File patterns listed for inclusion should be valid regular "
                "expressions"
            )

        QMessageBox.critical(self, _("Error"), msg)

    def is_valid(self):
        # Check regexs
        try:
            flake8_filenames_matches = (
                self.flake8_filenames_match.textbox.text().split(",")
            )
            for match in flake8_filenames_matches:
                re.compile(match.strip())
        except re.error:
            self.report_invalid_regex()
            return False

        try:
            # flake8 check
            flake8_excludes = self.flake8_exclude.textbox.text().split(",")
            for match in flake8_excludes:
                re.compile(match.strip())

            # ruff check
            ruff_excludes = self.ruff_exclude.textbox.text().split(",")
            for match in ruff_excludes:
                re.compile(match.strip())

        except re.error:
            self.report_invalid_regex(files=False)
            return False

        return True



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Shell Widget for the IPython Console
"""

# Standard library imports
from collections.abc import Callable
import logging
import os
import os.path as osp
import time
from textwrap import dedent
import typing

# Third party imports
from qtpy.QtCore import Qt, Signal, QEvent
from qtpy.QtGui import QClipboard, QTextCursor, QTextFormat
from qtpy.QtWidgets import QApplication, QMessageBox
from spyder_kernels.comms.frontendcomm import CommError
from spyder_kernels.utils.style import create_style_class
from traitlets import observe

# Local imports
from spyder.api.plugins import Plugins
from spyder.api.translations import _
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.config.base import is_conda_based_app, running_under_pytest
from spyder.config.gui import get_color_scheme, is_dark_interface
from spyder.plugins.ipythonconsole.api import (
    IPythonConsoleWidgetCornerWidgets,
    IPythonConsoleWidgetMenus,
    ClientContextMenuActions,
    ClientContextMenuSections
)
from spyder.plugins.ipythonconsole.utils.style import create_qss_style
from spyder.plugins.ipythonconsole.utils.kernel_handler import (
    KernelConnectionState)
from spyder.plugins.ipythonconsole.widgets import (
    ControlWidget, DebuggingWidget, FigureBrowserWidget, HelpWidget,
    NamepaceBrowserWidget, PageControlWidget)
from spyder.utils import syntaxhighlighters as sh
from spyder.utils.palette import SpyderPalette
from spyder.utils.clipboard_helper import CLIPBOARD_HELPER
from spyder.widgets.helperwidgets import MessageCheckBox

if typing.TYPE_CHECKING:
    from spyder.plugins.ipythonconsole.widgets.client import ClientWidget


logger = logging.getLogger(__name__)

MODULES_FAQ_URL = (
    "https://docs.spyder-ide.org/5/faq.html#using-packages-installer"
)


class ShellWidget(NamepaceBrowserWidget, HelpWidget, DebuggingWidget,
                  FigureBrowserWidget, SpyderWidgetMixin):
    """
    Shell widget for the IPython Console

    This is the widget in charge of executing code
    """
    PLUGIN_NAME = Plugins.IPythonConsole

    # NOTE: Signals can't be assigned separately to each widget
    #       That's why we define all needed signals here.

    # For NamepaceBrowserWidget
    sig_show_syspath = Signal(object)
    sig_show_env = Signal(object)

    # For FigureBrowserWidget
    sig_new_inline_figure = Signal(object, str)

    # For DebuggingWidget
    sig_pdb_step = Signal(str, int)
    """
    This signal is emitted when Pdb reaches a new line.

    Parameters
    ----------
    filename: str
        The filename the debugger stepped in.
    line_number: int
        The line number the debugger stepped in.
    """

    sig_pdb_stack = Signal(object, int)
    """
    This signal is emitted when the Pdb stack changed.

    Parameters
    ----------
    pdb_stack: traceback.StackSummary
        The current pdb stack.
    pdb_index: int
        The index in the stack.
    """

    sig_pdb_state_changed = Signal(bool)
    """
    This signal is emitted every time a Pdb interaction happens.

    Parameters
    ----------
    pdb_state: bool
        Whether the debugger is waiting for input.
    """

    sig_pdb_prompt_ready = Signal()
    """Called when pdb request new input"""

    # For ShellWidget
    sig_focus_changed = Signal()
    sig_new_client = Signal()

    # Kernel died and restarted (not user requested)
    sig_prompt_ready = Signal()
    sig_remote_execute = Signal()

    # For global working directory
    sig_working_directory_changed = Signal(str, str)

    # For printing internal errors
    sig_exception_occurred = Signal(dict)

    # To save values and messages returned by the kernel
    _kernel_is_starting = True

    # Request plugins to send additional configuration to the Spyder kernel
    sig_config_spyder_kernel = Signal()

    # Kernel ready
    sig_kernel_is_ready = Signal()
    """
    Signal used to inform other Qt objects that the kernel is ready.

    Notes
    -----
    * Do not connect directly to the `kernel_handler.sig_kernel_is_ready`
      signal to receive updates about this because that object is replaced on
      kernel restarts.
    * See spyder-ide/spyder#24577 for more context.
    """

    # To notify of kernel connection, disconnection and kernel errors
    sig_shellwidget_created = Signal(object)
    sig_shellwidget_deleted = Signal(object)
    sig_shellwidget_errored = Signal(object)

    # To request restart
    sig_restart_kernel = Signal()

    sig_kernel_state_arrived = Signal(dict)
    """
    A new kernel state, which needs to be processed.

    Parameters
    ----------
    state: dict
        Kernel state. The structure of this dictionary is defined in the
        `SpyderKernel.get_state` method of Spyder-kernels.
    """

    def __init__(
        self,
        ipyclient,
        additional_options,
        handlers,
        *args,
        special_kernel=None,
        **kw,
    ):
        # To override the Qt widget used by RichJupyterWidget
        self.custom_control = ControlWidget
        self.custom_page_control = PageControlWidget
        self.custom_edit = True

        super().__init__(*args, **kw)
        self.ipyclient: ClientWidget = ipyclient
        self.additional_options = additional_options
        self.special_kernel = special_kernel

        # Keyboard shortcuts
        # Registered here to use shellwidget as the parent
        SpyderWidgetMixin.__init__(self)
        self.regiter_shortcuts()

        # Set the color of the matched parentheses here since the qtconsole
        # uses a hard-coded value that is not modified when the color scheme is
        # set in the qtconsole constructor. See spyder-ide/spyder#4806.
        self.set_bracket_matcher_color_scheme(self.syntax_style)

        self.shutting_down = False
        self.kernel_manager = None
        self.kernel_client = None
        self.kernel_handler = None
        self._kernel_configuration = {}
        self.is_kernel_configured = False
        self._init_kernel_setup = False
        self._is_banner_shown = False

        # Set bright colors instead of bold formatting for better traceback
        # readability.
        self._ansi_processor.bold_text_enabled = False

        if handlers is None:
            handlers = {}
        else:
            # Avoid changing the plugin dict
            handlers = handlers.copy()
        handlers.update({
            'show_pdb_output': self.show_pdb_output,
            'pdb_input': self.pdb_input,
            'update_state': self.update_state,
        })
        self.kernel_comm_handlers = handlers

        # To keep an execution queue
        self._execute_queue = []
        self.executed.connect(self.pop_execute_queue)

        # Show a message in our installers to explain users how to use
        # modules that don't come with them.
        self.show_modules_message = is_conda_based_app()

        # The Qtconsole shortcuts for the actions below don't work in Spyder,
        # so we disable them.
        self._copy_raw_action.setShortcut('')
        self.export_action.setShortcut('')
        self.select_all_action.setShortcut('')
        self.print_action.setShortcut('')

    # ---- Public API
    @property
    def server_id(self):
        if self.is_remote():
            return self.ipyclient.jupyter_api.server_id

    @property
    def is_spyder_kernel(self):
        if self.kernel_handler is None:
            return False
        return self.kernel_handler.known_spyder_kernel

    @property
    def spyder_kernel_ready(self):
        """
        Check if Spyder kernel is ready.

        Notes
        -----
        This is used for our tests.
        """
        if self.kernel_handler is None:
            return False
        return (
            self.kernel_handler.connection_state ==
            KernelConnectionState.SpyderKernelReady)

    def connect_kernel(self, kernel_handler, first_connect=True):
        """Connect to the kernel using our handler."""
        # Kernel client
        kernel_client = kernel_handler.kernel_client
        kernel_client.stopped_channels.connect(self.notify_deleted)
        self.kernel_client = kernel_client

        self.kernel_manager = kernel_handler.kernel_manager
        self.kernel_handler = kernel_handler

        # Register handlers declared here before emitting sig_kernel_is_ready
        # so that handlers declared elsewhere can't be called first, which can
        # generate errors.
        for request_id, handler in self.kernel_comm_handlers.items():
            self.kernel_handler.kernel_comm.register_call_handler(
                request_id, handler
            )

        if first_connect:
            # Let plugins know that a new kernel is connected
            self.sig_shellwidget_created.emit(self)
        else:
            # Set _starting to False to avoid reset at first prompt
            self._starting = False

        # Connect signals
        kernel_handler.sig_kernel_is_ready.connect(
            self.handle_kernel_is_ready)
        kernel_handler.sig_kernel_connection_error.connect(
            self.handle_kernel_connection_error)

        kernel_handler.connect_()

    def disconnect_kernel(self, shutdown_kernel=True, will_reconnect=True):
        """
        Disconnect from current kernel.

        Parameters:
        -----------
        shutdown_kernel: bool
            If True, the kernel is shut down.
        will_reconnect: bool
            If False, emits `sig_shellwidget_deleted` so the plugins can close
            related widgets.
        """
        kernel_handler = self.kernel_handler
        if not kernel_handler:
            return
        kernel_client = kernel_handler.kernel_client

        kernel_handler.sig_kernel_is_ready.disconnect(
            self.handle_kernel_is_ready)
        kernel_handler.sig_kernel_connection_error.disconnect(
            self.handle_kernel_connection_error)
        kernel_handler.kernel_client.stopped_channels.disconnect(
            self.notify_deleted)

        if self._init_kernel_setup:
            self._init_kernel_setup = False

            kernel_handler.kernel_comm.sig_exception_occurred.disconnect(
                self.sig_exception_occurred)
            kernel_client.control_channel.message_received.disconnect(
                self._dispatch)

        kernel_handler.close(shutdown_kernel)
        if not will_reconnect:
            self.notify_deleted()
        # Reset state
        self.reset_kernel_state()

        self.kernel_client = None
        self.kernel_manager = None
        self.kernel_handler = None

    def handle_kernel_is_ready(self):
        """The kernel is ready"""
        if (
            self.kernel_handler.connection_state ==
            KernelConnectionState.SpyderKernelReady
        ):
            self.setup_spyder_kernel()
            self._show_banner()
            self.sig_kernel_is_ready.emit()

    def handle_kernel_connection_error(self):
        """An error occurred when connecting to the kernel."""
        if self.kernel_handler.connection_state == KernelConnectionState.Error:
            # A wrong version is connected
            self.ipyclient.show_kernel_error(
                self.kernel_handler.kernel_error_message,
            )

    def notify_deleted(self):
        """Notify that the shellwidget was deleted."""
        self.sig_shellwidget_deleted.emit(self)

    def shutdown(self, shutdown_kernel=True):
        """Shutdown connection and kernel."""
        if self.shutting_down:
            return
        self.shutting_down = True
        if self.kernel_handler is not None:
            self.kernel_handler.close(shutdown_kernel)
        super().shutdown()

    def reset_kernel_state(self):
        """Reset the kernel state."""
        self._prompt_requested = False
        self._pdb_recursion_level = 0
        self._reading = False

    def call_kernel(self, interrupt=False, blocking=False, callback=None,
                    timeout=None, display_error=False):
        """
        Send message to Spyder kernel connected to this console.

        Parameters
        ----------
        interrupt: bool
            Interrupt the kernel while running or in Pdb to perform
            the call.
        blocking: bool
            Make a blocking call, i.e. wait on this side until the
            kernel sends its response.
        callback: callable
            Callable to process the response sent from the kernel
            on the Spyder side.
        timeout: int or None
            Maximum time (in seconds) before giving up when making a
            blocking call to the kernel. If None, a default timeout
            (defined in commbase.py, present in spyder-kernels) is
            used.
        display_error: bool
            If an error occurs, should it be printed to the console.
        """
        return self.kernel_handler.kernel_comm.remote_call(
            interrupt=interrupt,
            blocking=blocking,
            callback=callback,
            timeout=timeout,
            display_error=display_error
        )

    @property
    def is_external_kernel(self):
        """Check if this is an external kernel."""
        return self.kernel_manager is None

    def setup_spyder_kernel(self):
        """Setup spyder kernel"""
        if not self._init_kernel_setup:
            # Only do this setup once
            self._init_kernel_setup = True

            # For errors
            self.kernel_handler.kernel_comm.sig_exception_occurred.connect(
                self.sig_exception_occurred)

            # For completions
            self.kernel_client.control_channel.message_received.connect(
                self._dispatch)

            # Redefine the complete method to work while debugging.
            self._redefine_complete_for_dbg(self.kernel_client)

        # Setup to do after restart
        # Check for fault and send config
        self.kernel_handler.poll_fault_text()

        self.send_spyder_kernel_configuration()

        # Update sys.path
        paths = self.get_conf(
            "spyder_pythonpath", section="pythonpath_manager"
        )
        prioritize = self.get_conf(
            "prioritize", section="pythonpath_manager"
        )
        self.update_syspath(paths, prioritize)

        run_lines = self.get_conf('startup/run_lines')
        if run_lines:
            self.execute(run_lines, hidden=True)

        if self.get_conf('startup/use_run_file'):
            run_file = self.get_conf('startup/run_file')
            if run_file:
                self.call_kernel().safe_exec(run_file)

    def send_spyder_kernel_configuration(self):
        """Send kernel configuration to spyder kernel."""
        self.is_kernel_configured = False

        # Set matplotlib backend
        self.send_mpl_backend()

        # set special kernel
        self.set_special_kernel()

        # Set current cwd
        self.set_cwd()

        # To apply style
        self.set_color_scheme(self.syntax_style, reset=False)

        # Enable faulthandler
        self.set_kernel_configuration("faulthandler", True)

        # Give a chance to plugins to configure the kernel
        self.sig_config_spyder_kernel.emit()

        if self.is_external_kernel:
            # Enable wurlitzer
            # Not necessary if started by spyder
            # Does not work if the external kernel is on windows
            self.set_kernel_configuration("wurlitzer", True)

        if self.get_conf('autoreload'):
            # Enable autoreload_magic
            self.set_kernel_configuration("autoreload_magic", True)

        self.call_kernel(
            interrupt=self.is_debugging(),
            callback=self.kernel_configure_callback
        ).set_configuration(self._kernel_configuration)

        self.is_kernel_configured = True

    def set_kernel_configuration(self, key, value):
        """Set kernel configuration."""
        if self.is_kernel_configured:
            if (
                key not in self._kernel_configuration
                or self._kernel_configuration[key] != value
            ):
                # Do not send twice
                self.call_kernel(
                    interrupt=self.is_debugging(),
                    callback=self.kernel_configure_callback
                ).set_configuration({key: value})

        self._kernel_configuration[key] = value

    def register_kernel_call_handler(self, handler_id: str, handler: Callable):
        """Register a comm handler."""
        self.kernel_comm_handlers[handler_id] = handler

        # When this class is initialized, not all handlers are passed to its
        # constructor (i.e. some are registered afterwards). But after a kernel
        # restart, all handlers should have been registered. So, this check
        # prevents registering non-init handlers multiple times after a restart
        if (
            handler_id
            not in self.kernel_handler.kernel_comm._remote_call_handlers
        ):
            self.kernel_handler.kernel_comm.register_call_handler(
                handler_id, handler
            )

    def unregister_kernel_call_handler(self, handler_id: str):
        """Unregister a comm handler."""
        self.kernel_comm_handlers.pop(handler_id, None)
        self.kernel_handler.kernel_comm.unregister_call_handler(handler_id)

    def kernel_configure_callback(self, dic):
        """Kernel configuration callback"""
        for key, value in dic.items():
            if key == "faulthandler":
                self.kernel_handler.faulthandler_setup(value)
            elif key == "special_kernel_error":
                self.ipyclient._show_special_console_error(value)

    def pop_execute_queue(self):
        """Pop one waiting instruction."""
        if self._execute_queue:
            self.execute(*self._execute_queue.pop(0))

    def interrupt_kernel(self):
        """Attempts to interrupt the running kernel."""
        # Empty queue when interrupting
        # Fixes spyder-ide/spyder#7293.
        self._execute_queue = []

        if self.spyder_kernel_ready:
            self._reading = False

            # Check if there is a kernel that can be interrupted before trying
            # to do it.
            # Fixes spyder-ide/spyder#20212
            if self.kernel_manager and self.kernel_manager.has_kernel:
                self.call_kernel(interrupt=True).raise_interrupt_signal()
            elif self.is_remote():
                # Request an interrupt to the server for remote kernels
                self.ipyclient.interrupt_remote_kernel()
            else:
                self._append_html(
                    _("<br><br>The kernel appears to be dead, so it can't be "
                      "interrupted. Please open a new console to keep "
                      "working.<br>")
                )
        else:
            self._append_html(
                _("<br><br>It is not possible to interrupt a non-Spyder "
                  "kernel I did not start.<br>")
            )

    def execute(self, source=None, hidden=False, interactive=False):
        """
        Executes source or the input buffer, possibly prompting for more
        input.
        """
        # Start remote wait indicator for network loss
        if self.is_remote():
            self.ipyclient._start_execution_loading_timer()

        # Needed for cases where there is no kernel initialized but
        # an execution is triggered like when setting initial configs.
        # See spyder-ide/spyder#16896
        if self.kernel_client is None:
            return
        if self._executing:
            self._execute_queue.append((source, hidden, interactive))
            return
        super().execute(source, hidden, interactive)

    def is_running(self):
        """Check if shell is running."""
        return (
            self.kernel_client is not None and
            self.kernel_client.channels_running
        )

    def set_cwd(self, dirname=None, emit_cwd_change=False):
        """
        Set shell current working directory.

        Parameters
        ----------
        dirname: str
            Path to the new current working directory.
        emit_cwd_change: bool
            Whether to emit a Qt signal that informs other panes in Spyder that
            the current working directory has changed.
        """
        if dirname is None:
            if not self.get_cwd():
                return
            dirname = self.get_cwd()
        elif os.name == 'nt' and self.server_id is None:
            # Use normpath instead of replacing '\' with '\\'
            # See spyder-ide/spyder#10785
            dirname = osp.normpath(dirname)

        logger.debug(f"Attempt to change cwd to {dirname}")
        self.set_kernel_configuration("cwd", dirname)

        if emit_cwd_change:
            self.sig_working_directory_changed.emit(dirname, self.server_id)

    def send_mpl_backend(self, option=None):
        """
        Send matplotlib backend.

        If `option` is not None only send the related options.
        """
        if not self.spyder_kernel_ready:
            # will be sent later
            return

        # Set Matplotlib backend with Spyder options
        pylab_n = 'pylab'
        pylab_o = self.get_conf(pylab_n)

        if option is not None and not pylab_o:
            # The options are only related to pylab_o
            # So no need to change the backend
            return

        pylab_autoload_n = 'pylab/autoload'
        pylab_backend_n = 'pylab/backend'
        figure_format_n = 'pylab/inline/figure_format'
        resolution_n = 'pylab/inline/resolution'
        width_n = 'pylab/inline/width'
        height_n = 'pylab/inline/height'
        fontsize_n = 'pylab/inline/fontsize'
        bottom_n = 'pylab/inline/bottom'
        bbox_inches_n = 'pylab/inline/bbox_inches'
        backend_o = self.get_conf(pylab_backend_n)

        inline_backend = 'inline'
        matplotlib_conf = {}

        if pylab_o:
            # Figure format
            format_o = self.get_conf(figure_format_n)
            if format_o and (option is None or figure_format_n in option):
                matplotlib_conf[figure_format_n] = format_o

            # Resolution
            resolution_o = self.get_conf(resolution_n)
            if resolution_o is not None and (
                    option is None or resolution_n in option):
                matplotlib_conf[resolution_n] = resolution_o

            # Figure size
            width_o = float(self.get_conf(width_n))
            height_o = float(self.get_conf(height_n))
            if option is None or (width_n in option or height_n in option):
                if width_o is not None:
                    matplotlib_conf[width_n] = width_o
                if height_o is not None:
                    matplotlib_conf[height_n] = height_o

            # Font size
            fontsize_o = float(self.get_conf(fontsize_n))
            if (
                fontsize_o is not None
                and (option is None or fontsize_n in option)
            ):
                matplotlib_conf[fontsize_n] = fontsize_o

            # Bottom part
            bottom_o = float(self.get_conf(bottom_n))
            if (
                bottom_o is not None
                and (option is None or bottom_n in option)
            ):
                matplotlib_conf[bottom_n] = bottom_o

            # Print figure kwargs
            bbox_inches_o = self.get_conf(bbox_inches_n)
            if option is None or bbox_inches_n in option:
                matplotlib_conf[bbox_inches_n] = bbox_inches_o

        if pylab_o and backend_o is not None:
            mpl_backend = backend_o
        else:
            # Set Matplotlib backend to inline for external kernels.
            # Fixes issue spyder-ide/spyder-kernels#108
            mpl_backend = inline_backend

        # Automatically load Pylab and Numpy, or only set Matplotlib
        # backend
        autoload_pylab_o = self.get_conf(pylab_autoload_n)
        if option is None or pylab_backend_n in option:
            matplotlib_conf[pylab_backend_n] = mpl_backend
        if option is None or pylab_autoload_n in option:
            matplotlib_conf[pylab_autoload_n] = autoload_pylab_o

        if matplotlib_conf and pylab_o:
            self.set_kernel_configuration("matplotlib", matplotlib_conf)

    def get_cwd(self):
        """
        Get current working directory.

        Notes
        -----
        * This doesn't ask the kernel for its working directory. Instead, it
          returns the last value of it saved here.
        * We do it for performance reasons because we call this method when
          switching consoles to update the Working Directory toolbar.
        """
        return self._kernel_configuration.get("cwd", '')

    def update_state(self, state):
        """
        New state received from kernel.
        """
        cwd = state.pop("cwd", None)
        if cwd and cwd != self.get_cwd():
            # Only set it if self.get_cwd() is already set
            self._kernel_configuration["cwd"] = cwd
            self.sig_working_directory_changed.emit(cwd, self.server_id)

        if state:
            self.sig_kernel_state_arrived.emit(state)

    def set_bracket_matcher_color_scheme(self, color_scheme):
        """Set color scheme for matched parentheses."""
        bsh = sh.BaseSH(parent=self, color_scheme=color_scheme)
        mpcolor = bsh.get_matched_p_color()
        self._bracket_matcher.format.setBackground(mpcolor)

    def set_color_scheme(self, color_scheme, reset=True):
        """Set color scheme of the shell."""
        self.set_bracket_matcher_color_scheme(color_scheme)
        self.style_sheet, dark_color = create_qss_style(color_scheme)
        self.syntax_style = color_scheme

        if not self.spyder_kernel_ready:
            # Will be sent later
            return

        if reset:
            # Don't clear console and show a message instead to prevent
            # removing important content from users' consoles.
            # Fixes spyder-ide/spyder#9896
            self.reset(clear=False)
            self._append_plain_text(
                "\n\nNote: Clearing the console is necessary to fully apply "
                "the new syntax style you selected."
            )
        self.set_kernel_configuration(
            "color scheme", "dark" if not dark_color else "light"
        )
        color_scheme = get_color_scheme(self.syntax_style)
        self.set_kernel_configuration(
            "traceback_highlight_style",
            color_scheme,
        )

    def update_syspath(self, new_paths, prioritize):
        """Update sys.path contents in the kernel."""
        # Prevent error when the kernel is not available and users open/close
        # projects or use the Python path manager.
        # Fixes spyder-ide/spyder#21563
        if self.kernel_handler is not None:
            self.call_kernel(interrupt=True, blocking=False).update_syspath(
                new_paths, prioritize
            )

    def request_syspath(self):
        """Ask the kernel for sys.path contents."""
        self.call_kernel(
            interrupt=True, callback=self.sig_show_syspath.emit).get_syspath()

    def request_env(self):
        """Ask the kernel for environment variables."""
        self.call_kernel(
            interrupt=True, callback=self.sig_show_env.emit).get_env()

    def set_show_calltips(self, show_calltips):
        """Enable/Disable showing calltips."""
        self.enable_calltips = show_calltips

    def set_buffer_size(self, buffer_size):
        """Set buffer size for the shell."""
        self.buffer_size = buffer_size

    def set_completion_type(self, completion_type):
        """Set completion type (Graphical, Terminal, Plain) for the shell."""
        self.gui_completion = completion_type

    def set_in_prompt(self, in_prompt):
        """Set appereance of the In prompt."""
        self.in_prompt = in_prompt

    def set_out_prompt(self, out_prompt):
        """Set appereance of the Out prompt."""
        self.out_prompt = out_prompt

    def get_matplotlib_backend(self):
        """Call kernel to get current backend."""
        return self.call_kernel(
            interrupt=True,
            blocking=True).get_matplotlib_backend()

    def get_mpl_interactive_backend(self):
        """Call kernel to get current interactive backend."""
        return self.call_kernel(
            interrupt=True,
            blocking=True).get_mpl_interactive_backend()

    def set_jedi_completer(self, use_jedi):
        """Set if jedi completions should be used."""
        self.set_kernel_configuration(
            "jedi_completer", use_jedi
        )

    def set_greedy_completer(self, use_greedy):
        """Set if greedy completions should be used."""
        self.set_kernel_configuration(
            "greedy_completer", use_greedy
        )

    def set_autocall(self, autocall):
        """Set if autocall functionality is enabled or not."""
        self.set_kernel_configuration(
            "autocall", autocall
        )

    # --- To handle the banner
    def long_banner(self):
        """Banner for clients with additional content."""
        # Default banner
        try:
            env_info = self.get_pythonenv_info()
            sys_version = env_info['sys_version'].replace('\n', '')
            ipython_version = env_info["ipython_version"]

            banner_parts = [
                f"Python {sys_version}\n",
                'Type "copyright", "credits" or "license" for more '
                "information.",
                "\n\n",
            ]

            banner_parts.append(
                f"IPython {ipython_version} -- An enhanced Interactive "
                f"Python. Type '?' for help.\n"
            )

            banner = ''.join(banner_parts)
        except (CommError, TimeoutError, RuntimeError):
            # RuntimeError happens when the kernel crashes after it starts.
            # See spyder-ide/spyder#22929
            banner = ""

        # Pylab additions
        pylab_o = self.additional_options['pylab']
        autoload_pylab_o = self.additional_options['autoload_pylab']
        if pylab_o and autoload_pylab_o:
            pylab_message = ("\nPopulating the interactive namespace from "
                             "numpy and matplotlib\n")
            banner = banner + pylab_message

        # Sympy additions
        sympy_o = self.additional_options['sympy']
        if sympy_o:
            lines = """
These commands were executed:
>>> from sympy import *
>>> x, y, z, t = symbols('x y z t')
>>> k, m, n = symbols('k m n', integer=True)
>>> f, g, h = symbols('f g h', cls=Function)
"""
            banner = banner + lines
        if (pylab_o and sympy_o):
            lines = """
Warning: Pylab (i.e. Numpy and Matplotlib) and symbolic math (Sympy) are both
enabled at the same time. Hence, some Matplotlib functions are going to be
overrided by the Sympy module (e.g. plot)
"""
            banner = banner + lines

        return banner

    def short_banner(self):
        """Short banner with Python and IPython versions only."""
        try:
            env_info = self.get_pythonenv_info()
            py_ver = env_info['python_version']
            ipy_ver = env_info['ipython_version']
            banner = f'Python {py_ver} -- IPython {ipy_ver}\n'
        except (CommError, TimeoutError):
            banner = ""

        return banner

    # --- To define additional shortcuts
    def clear_console(self):
        self.execute("%clear")
        # Stop reading as any input has been removed.
        self._reading = False

    def reset_namespace(self, warning=False, message=False):
        """Reset the namespace by removing all names defined by the user."""
        # Don't show the warning when running our tests.
        if running_under_pytest():
            warning = False

        if warning:
            reset_str = _("Remove all variables")
            warn_str = _("All user-defined variables will be removed. "
                         "Are you sure you want to proceed?")
            box = MessageCheckBox(icon=QMessageBox.Warning, parent=self)
            box.setWindowTitle(reset_str)
            box.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
            box.setDefaultButton(QMessageBox.Yes)

            box.set_checkbox_text(_("Don't show again."))
            box.set_checked(False)
            box.set_check_visible(True)
            box.setText(warn_str)

            box.buttonClicked.connect(
                lambda button: self.handle_reset_message_answer(
                    box, button, message)
            )
            box.show()
        else:
            self._perform_reset(message)

    def handle_reset_message_answer(self, message_box, button, message):
        """
        Handle the answer of the reset namespace message box.

        Parameters
        ----------
        message_box
            Instance of the message box shown to the user.
        button: QPushButton
            Instance of the button clicked by the user on the dialog.
        message: bool
            Whether to show a message in the console telling users the
            namespace was reset.
        """
        if message_box.buttonRole(button) == QMessageBox.YesRole:
            self._update_reset_options(message_box)
            self._perform_reset(message)
        else:
            self._update_reset_options(message_box)

    def set_special_kernel(self):
        """Reset special kernel"""
        if not self.special_kernel:
            return

        # Check if the dependecies for special consoles are available.
        self.set_kernel_configuration(
            "special_kernel", self.special_kernel
        )

    def regiter_shortcuts(self):
        """Register shortcuts for this widget."""

        shortcuts = (
            ('Inspect current object', self._control.inspect_current_object),
            ('Clear shell', self.clear_console),
            ('Restart kernel', self.sig_restart_kernel),
            ('new tab', self.sig_new_client),
            ('reset namespace', self._reset_namespace),
            ('enter array inline', self._control.enter_array_inline),
            ('enter array table', self._control.enter_array_table),
            ('clear line', self.ipyclient.clear_line),
        )

        for name, callback in shortcuts:
            self.register_shortcut_for_widget(name=name, triggered=callback)

    # --- To communicate with the kernel
    def silent_execute(self, code):
        """Execute code in the kernel without increasing the prompt"""
        try:
            if self.is_debugging():
                self.pdb_execute(code, hidden=True)
            else:
                self.kernel_client.execute(str(code), silent=True)
        except AttributeError:
            pass

    def set_backend_for_mayavi(self, command):
        """
        Mayavi plots require the Qt backend, so we try to detect if one is
        generated to change backends
        """
        calling_mayavi = False
        lines = command.splitlines()
        for line in lines:
            if not line.startswith('#'):
                if 'import mayavi' in line or 'from mayavi' in line:
                    calling_mayavi = True
                    break
        if calling_mayavi:
            message = _("Changing backend to Qt for Mayavi")
            self._append_plain_text(message + '\n')
            self.silent_execute("%gui inline\n%gui qt")

    def change_mpl_backend(self, command):
        """
        If the user is trying to change Matplotlib backends with
        %matplotlib, send the same command again to the kernel to
        correctly change it.

        Fixes spyder-ide/spyder#4002.
        """
        if (command.startswith('%matplotlib') and
                len(command.splitlines()) == 1):
            if 'inline' not in command:
                self.silent_execute(command)

    def append_html_message(self, html, before_prompt=False,
                            msg_type='warning'):
        """
        Append an html message enclosed in a box.

        Parameters
        ----------
        before_prompt: bool
            Whether to add the message before the next prompt.
        msg_type: str
            Type of message to be showm. Possible values are
            'warning' and 'error'.
        """
        # The message is displayed in a table with a header and a single cell.
        table_properties = (
            "border='0.5'" +
            "width='90%'" +
            "cellpadding='8'" +
            "cellspacing='0'"
        )

        if msg_type == 'error':
            header = _("Error")
            bgcolor = SpyderPalette.COLOR_ERROR_2
        else:
            header = _("Important")
            bgcolor = SpyderPalette.COLOR_WARN_1

        # This makes the header text have good contrast against its background
        # for the light theme.
        if is_dark_interface():
            font_color = SpyderPalette.COLOR_TEXT_1
        else:
            font_color = 'white'

        self._append_html(
            f"<div align='center'>"
            f"<table {table_properties}>"
            # Header
            f"<tr><th bgcolor='{bgcolor}'><font color='{font_color}'>"
            f"{header}"
            f"</th></tr>"
            # Cell with html message
            f"<tr><td>{html}</td></tr>"
            f"</table>"
            f"</div>",
            before_prompt=before_prompt
        )

    def insert_horizontal_ruler(self):
        """
        Insert a horizontal ruler at the current cursor position.

        Notes
        -----
        This only works when adding a single horizontal line to a
        message. For more complex messages, please use
        append_html_message.
        """
        self._control.insert_horizontal_ruler()

    def get_pythonenv_info(self):
        """Call kernel to get the current Python environment info."""
        return self.call_kernel(
            interrupt=True, blocking=True
        ).get_pythonenv_info()

    def is_remote(self):
        """Check if this shell is connected to a remote server."""
        return self.ipyclient.is_remote()

    # ---- Public methods (overrode by us)
    def paste(self, mode=QClipboard.Clipboard):
        """ Paste the contents of the clipboard into the input region.

        Parameters
        ----------
        mode : QClipboard::Mode, optional [default QClipboard::Clipboard]

            Controls which part of the system clipboard is used. This can be
            used to access the selection clipboard in X11 and the Find buffer
            in Mac OS. By default, the regular clipboard is used.
        """
        if self._control.textInteractionFlags() & Qt.TextEditable:
            # Make sure the paste is safe.
            self._keep_cursor_in_buffer()
            cursor = self._control.textCursor()

            # Remove any trailing newline, which confuses the GUI and forces
            # the user to backspace.
            text = QApplication.clipboard().text(mode).rstrip()

            # Adjust indentation of multilines pastes
            if len(text.splitlines()) > 1:
                lines_adjustment = CLIPBOARD_HELPER.remaining_lines_adjustment(
                    self._get_preceding_text())
                eol_chars = "\n"
                first_line, *remaining_lines = (text + eol_chars).splitlines()
                remaining_lines = [
                    self._adjust_indentation(line, lines_adjustment)
                    for line in remaining_lines]
                text = eol_chars.join([first_line, *remaining_lines])

            # dedent removes "common leading whitespace" but to preserve
            # relative indent of multiline code, we have to compensate for any
            # leading space on the first line, if we're pasting into
            # an indented position.
            cursor_offset = cursor.position() - self._get_line_start_pos()
            if text.startswith(' ' * cursor_offset):
                text = text[cursor_offset:]

            self._insert_plain_text_into_buffer(cursor, dedent(text))

    def copy(self):
        """
        Copy the currently selected text to the clipboard.
        """
        super().copy()
        self._save_clipboard_indentation()

    def cut(self):
        """
        Copy the currently selected text to the clipboard and delete it
        if it's inside the input buffer.
        """
        super().cut()
        self._save_clipboard_indentation()

    # ---- Private API
    def _adjust_indentation(self, line, indent_adjustment):
        """Adjust indentation."""
        if indent_adjustment == 0 or line == "":
            return line

        if indent_adjustment > 0:
            return ' ' * indent_adjustment + line

        max_indent = CLIPBOARD_HELPER.get_line_indentation(line)
        indent_adjustment = min(max_indent, -indent_adjustment)

        return line[indent_adjustment:]

    def _get_preceding_text(self):
        """Get preciding text."""
        cursor = self._control.textCursor()
        text = cursor.selection().toPlainText()
        if text == "":
            return ""
        first_line_selection = text.splitlines()[0]
        cursor.setPosition(cursor.selectionStart())
        cursor.setPosition(cursor.block().position(), QTextCursor.KeepAnchor)
        preceding_text = cursor.selection().toPlainText()
        first_line = preceding_text + first_line_selection
        len_with_prompt = len(first_line)
        # Remove prompt
        first_line = self._highlighter.transform_classic_prompt(first_line)
        first_line = self._highlighter.transform_ipy_prompt(first_line)

        prompt_len = len_with_prompt - len(first_line)
        if prompt_len >= len(preceding_text):
            return ""

        return preceding_text[prompt_len:]

    def _save_clipboard_indentation(self):
        """
        Save the indentation corresponding to the clipboard data.

        Must be called right after copying.
        """
        CLIPBOARD_HELPER.save_indentation(self._get_preceding_text(), 4)

    def _show_banner(self):
        """Show banner before first prompt."""
        if (
            # Don't show banner for external but local kernels
            self.is_external_kernel and not self.is_remote()
            # Don't show it if it was already shown
            or self._is_banner_shown
        ):
            return

        logger.debug(f"Showing banner for {self}")

        # Check what kind of banner we want to show
        show_banner_o = self.additional_options['show_banner']
        if show_banner_o:
            banner = self.long_banner()
        else:
            banner = self.short_banner()

        # Move cursor to first position and insert banner
        cursor = self._control.textCursor()
        cursor.setPosition(0)
        self._insert_plain_text(cursor, banner)

        # We need to do this so the banner is available to other QtConsole
        # methods (e.g. console resets).
        # Fixes spyder-ide/spyder#22593
        self.banner = banner

        # Ensure banner can not be undone after insertion
        # See https://github.com/spyder-ide/spyder/pull/24726#discussion_r2229475275
        self._control.document().clearUndoRedoStacks()

        # Only do this once
        self._is_banner_shown = True

    def _reset_namespace(self):
        warning = self.get_conf('show_reset_namespace_warning')
        self.reset_namespace(warning=warning)

    def _perform_reset(self, message):
        """
        Perform the reset namespace operation.

        Parameters
        ----------
        message: bool
            Whether to show a message in the console telling users the
            namespace was reset.
        """
        try:
            if self.is_waiting_pdb_input():
                self.execute('%reset -f')
            else:
                if message:
                    self.reset()
                    self._append_html(
                        _("<br><br>Removing all variables...<br>"),
                        before_prompt=False
                    )
                    self.insert_horizontal_ruler()
                self.silent_execute("%reset -f")
                self.set_special_kernel()

                if self.spyder_kernel_ready:
                    self.call_kernel().close_all_mpl_figures()
                    self.send_spyder_kernel_configuration()
        except AttributeError:
            pass

    def _update_reset_options(self, message_box):
        """
        Update options and variables based on the interaction in the
        reset warning message box shown to the user.
        """
        self.set_conf(
            'show_reset_namespace_warning',
            not message_box.is_checked()
        )
        self.ipyclient.reset_warning = not message_box.is_checked()

    # ---- Private API (overrode by us)
    def _event_filter_console_keypress(self, event):
        """Filter events to send to qtconsole code."""
        key = event.key()
        if self._control_key_down(event.modifiers(), include_command=False):
            if key == Qt.Key_Period:
                # Do not use ctrl + . to restart kernel
                # Handled by IPythonConsoleWidget
                return False
        return super()._event_filter_console_keypress(event)

    def _handle_execute_reply(self, msg):
        """
        Reimplemented to handle communications between Spyder
        and the kernel
        """
        # Notify that kernel has started
        exec_count = msg['content'].get('execution_count', '')
        if exec_count == 0 and self._kernel_is_starting:
            self.ipyclient.t0 = time.monotonic()
            self._kernel_is_starting = False

        if self.is_remote():
            self.ipyclient._stop_execution_loading()

        # This catches an error when doing the teardown of a test.
        try:
            super()._handle_execute_reply(msg)
        except RuntimeError:
            pass

    def _handle_status(self, msg):
        """
        Reimplemented to refresh the namespacebrowser after kernel
        restarts
        """
        state = msg['content'].get('execution_state', '')
        msg_type = msg['parent_header'].get('msg_type', '')
        if state == 'starting':
            # This is needed to show the time a kernel
            # has been alive in each console.
            self.ipyclient.t0 = time.monotonic()
            self.ipyclient.timer.timeout.connect(self.ipyclient.show_time)
            self.ipyclient.timer.start(1000)

            # This handles restarts when the kernel dies
            # unexpectedly
            if not self._kernel_is_starting:
                self._kernel_is_starting = True
        elif state == 'idle' and msg_type == 'shutdown_request':
            # This handles restarts asked by the user
            self.ipyclient.t0 = time.monotonic()
        else:
            super()._handle_status(msg)

    def _handle_error(self, msg):
        """
        Reimplemented to reset the prompt if the error comes after the reply
        """
        self._process_execute_error(msg)

    def _context_menu_make(self, pos):
        """Reimplement the Qtconsole context menu using our API for menus."""
        context_menu = self.get_menu(
            IPythonConsoleWidgetMenus.ClientContextMenu
        )
        context_menu.clear_actions()

        fmt = self._control.cursorForPosition(pos).charFormat()
        img_name = fmt.stringProperty(QTextFormat.ImageName)

        if img_name:
            # Add image/svg actions to menu
            for name in [ClientContextMenuActions.CopyImage,
                         ClientContextMenuActions.SaveImage]:
                action = self.get_action(name)
                action.setData(img_name)
                self.add_item_to_menu(
                    action,
                    context_menu,
                    section=ClientContextMenuSections.Image
                )

            svg = self._name_to_svg_map.get(img_name, None)
            if svg is not None:
                for name in [ClientContextMenuActions.CopySvg,
                             ClientContextMenuActions.SaveSvg]:
                    action = self.get_action(name)
                    action.setData(svg)
                    self.add_item_to_menu(
                        action,
                        context_menu,
                        section=ClientContextMenuSections.SVG
                    )
        else:
            # Enable/disable edit actions
            cut_action = self.get_action(ClientContextMenuActions.Cut)
            cut_action.setEnabled(self.can_cut())

            for name in [ClientContextMenuActions.Copy,
                         ClientContextMenuActions.CopyRaw]:
                action = self.get_action(name)
                action.setEnabled(self.can_copy())

            paste_action = self.get_action(ClientContextMenuActions.Paste)
            paste_action.setEnabled(self.can_paste())

            # Add regular actions to menu
            for name in [ClientContextMenuActions.Cut,
                         ClientContextMenuActions.Copy,
                         ClientContextMenuActions.CopyRaw,
                         ClientContextMenuActions.Paste,
                         ClientContextMenuActions.SelectAll]:
                self.add_item_to_menu(
                    self.get_action(name),
                    context_menu,
                    section=ClientContextMenuSections.Edit
                )

            self.add_item_to_menu(
                self.get_action(ClientContextMenuActions.InspectObject),
                context_menu,
                section=ClientContextMenuSections.Inspect
            )

            for name in [ClientContextMenuActions.ArrayTable,
                         ClientContextMenuActions.ArrayInline]:
                self.add_item_to_menu(
                    self.get_action(name),
                    context_menu,
                    section=ClientContextMenuSections.Array
                )

            for name in [ClientContextMenuActions.Export,
                         ClientContextMenuActions.Print]:
                self.add_item_to_menu(
                    self.get_action(name),
                    context_menu,
                    section=ClientContextMenuSections.Export
                )

            for name in [ClientContextMenuActions.ClearConsole,
                         ClientContextMenuActions.ClearLine]:
                self.add_item_to_menu(
                    self.get_action(name),
                    context_menu,
                    section=ClientContextMenuSections.Clear
                )

            self.add_item_to_menu(
                self.get_action(ClientContextMenuActions.Quit),
                context_menu,
                section=ClientContextMenuSections.Quit
            )

        return context_menu

    def _banner_default(self):
        """Override banner creation to handle it in Spyder."""
        return ""

    def _handle_kernel_died(self, since_last_heartbeat):
        """Handle the kernel's death (if we do not own the kernel)."""
        # Disable stop button
        stop_button = self.get_toolbutton(
            IPythonConsoleWidgetCornerWidgets.InterruptButton
        )
        stop_button.setEnabled(False)

        super()._handle_kernel_died(since_last_heartbeat)

    def _kernel_restarted_message(self, died=True, reconnected=False):
        if died:
            msg = _("The kernel died, restarting...")
        elif reconnected:
            msg = _("Kernel reconnected...")
        else:
            msg = _("Restarting kernel...")

        if (
            died
            and self.kernel_manager is None
            and not self.is_remote()
        ):
            # The kernel might never restart, show position of fault file
            # if available else show kernel error
            if self.kernel_handler.fault_filename():
                msg += (
                    "\n" + _("Its crash file is located at:") + " "
                    + self.kernel_handler.fault_filename()
                )
            else:
                self.ipyclient.show_kernel_connection_error()

        self._append_html(f"<br>{msg}<br>", before_prompt=False)
        self.insert_horizontal_ruler()

    def _handle_kernel_restarted(self, *args, **kwargs):
        """The kernel restarted."""
        super()._handle_kernel_restarted(*args, **kwargs)

        # Reset Pdb state
        self.reset_kernel_state()

        # reset comm
        self.kernel_handler.reopen_comm()

        # In case anyone waits on end of execution
        self.executed.emit({})

    @observe('syntax_style')
    def _syntax_style_changed(self, changed=None):
        """Refresh the highlighting with the current syntax style by class."""
        if self._highlighter is None:
            # ignore premature calls
            return
        if self.syntax_style:
            color_scheme = get_color_scheme(self.syntax_style)
            self._highlighter._style = create_style_class(color_scheme)
            self._highlighter._clear_caches()
        else:
            self._highlighter.set_style_sheet(self.style_sheet)

    def _get_color(self, color):
        """
        Get a color as qtconsole.styles._get_color() would return from
        a builtin Pygments style.
        """
        color_scheme = get_color_scheme(self.syntax_style)
        return dict(
            bgcolor=color_scheme['background'],
            select=color_scheme['background'],
            fgcolor=color_scheme['normal'][0])[color]

    def _prompt_started_hook(self):
        """Emit a signal when the prompt is ready."""
        if not self._reading:
            self._highlighter.highlighting_on = True
            self.sig_prompt_ready.emit()

    def _handle_execute_input(self, msg):
        """Handle an execute_input message"""
        super()._handle_execute_input(msg)
        self.sig_remote_execute.emit()

    def _process_execute_error(self, msg):
        """
        Display a message when using our installers to explain users
        how to use modules that doesn't come with them.
        """
        super()._process_execute_error(msg)
        if self.show_modules_message:
            error = msg['content']['traceback']
            if any(['ModuleNotFoundError' in frame or 'ImportError' in frame
                    for frame in error]):
                self.append_html_message(
                    _("It seems you're trying to use a module that doesn't "
                      "come with our installer. Check "
                      "<a href='{}'>this FAQ</a> in our docs to learn how "
                      "to do this.").format(MODULES_FAQ_URL),
                    before_prompt=True
                )
            self.show_modules_message = False

    # ---- Qt methods
    def focusInEvent(self, event):
        """Reimplement Qt method to send focus change notification"""
        self.sig_focus_changed.emit()
        return super().focusInEvent(event)

    def focusOutEvent(self, event):
        """Reimplement Qt method to send focus change notification"""
        self.sig_focus_changed.emit()
        return super().focusOutEvent(event)
    
    def eventFilter(self, obj, event):
        if (
            event.type() == QEvent.Wheel
            and self._control_key_down(event.modifiers())
            and self.get_conf('disable_zoom_mouse', section='main')
        ):
            return False

        return super().eventFilter(obj, event)

    # ---- Python methods
    def __repr__(self):
        # Handy repr for logging.
        # Solution from https://stackoverflow.com/a/121508/438386
        return (
            "<" + self.__class__.__name__ + " object at " + hex(id(self)) + ">"
        )



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# Standard library imports
import sys

# Third party imports
from qtpy.QtCore import Qt, QPoint
from qtpy.QtWidgets import QListWidget, QListWidgetItem

# Local imports
from spyder.api.shortcuts import SpyderShortcutsMixin
from spyder.utils.icon_manager import ima


class TabSwitcherWidget(QListWidget, SpyderShortcutsMixin):
    """Show tabs in mru order and change between them."""

    CONF_SECTION = "editor"

    def __init__(self, parent, stack_history, tabs):
        QListWidget.__init__(self, parent)
        self.setWindowFlags(Qt.FramelessWindowHint | Qt.Dialog)

        self.editor = parent
        self.stack_history = stack_history
        self.tabs = tabs

        self.setSelectionMode(QListWidget.SingleSelection)
        self.itemActivated.connect(self.item_selected)

        self.id_list = []
        self.load_data()
        size = self.get_conf('completion/size', section='main')
        self.resize(*size)
        self.set_dialog_position()
        self.setCurrentRow(0)

        self.register_shortcut_for_widget(
            name='Go to previous file',
            triggered=lambda: self.select_row(-1),
            context='Editor',
        )
        self.register_shortcut_for_widget(
            name='Go to next file',
            triggered=lambda: self.select_row(1),
            context='Editor',
        )

    def load_data(self):
        """Fill ListWidget with the tabs texts.

        Add elements in inverse order of stack_history.
        """
        for index in reversed(self.stack_history):
            text = self.tabs.tabText(index)
            text = text.replace('&', '')
            item = QListWidgetItem(ima.icon('TextFileIcon'), text)
            self.addItem(item)

    def item_selected(self, item=None):
        """Change to the selected document and hide this widget."""
        if item is None:
            item = self.currentItem()

        # stack history is in inverse order
        try:
            index = self.stack_history[-(self.currentRow()+1)]
        except IndexError:
            pass
        else:
            self.editor.set_stack_index(index)
            self.editor.current_changed(index)
        self.hide()

    def select_row(self, steps):
        """Move selected row a number of steps.

        Iterates in a cyclic behaviour.
        """
        row = (self.currentRow() + steps) % self.count()
        self.setCurrentRow(row)

    def set_dialog_position(self):
        """Positions the tab switcher in the top-center of the editor."""
        left = int(self.editor.geometry().width()/2 - self.width()/2)
        top = int(self.editor.tabs.tabBar().geometry().height() +
                  self.editor.fname_label.geometry().height())

        self.move(self.editor.mapToGlobal(QPoint(left, top)))

    def keyReleaseEvent(self, event):
        """Reimplement Qt method.

        Handle "most recent used" tab behavior,
        When ctrl is released and tab_switcher is visible, tab will be changed.
        """
        if self.isVisible():
            qsc = self.get_shortcut(name='go to next file', context='editor')

            for key in qsc.split('+'):
                key = key.lower()
                if ((key == 'ctrl' and event.key() == Qt.Key_Control) or
                        (key == 'alt' and event.key() == Qt.Key_Alt)):
                    self.item_selected()
        event.accept()

    def keyPressEvent(self, event):
        """Reimplement Qt method to allow cyclic behavior."""
        if event.key() == Qt.Key_Down:
            self.select_row(1)
        elif event.key() == Qt.Key_Up:
            self.select_row(-1)

    def focusOutEvent(self, event):
        """Reimplement Qt method to close the widget when loosing focus."""
        event.ignore()
        if sys.platform == "darwin":
            if event.reason() != Qt.ActiveWindowFocusReason:
                self.close()
        else:
            self.close()





# -----------------------------------------------------------------------------
# Copyright (c) 2023- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Helper classes to get and set the fonts used in Spyder.
"""

# Third-party imports
from qtpy.QtGui import QFont

# Local imports
from spyder.config.gui import get_font


class SpyderFontType:
    """
    Font types used in Spyder plugins and the entire application.

    This enum is meant to be used to get the :class:`QFont` object
    corresponding to each type. These are:

    * :attr:`Monospace` is used in the :guilabel:`Editor`,
      :guilabel:`IPython Console` and :guilabel:`History` panes
    * :attr:`Interface` is used by the entire Spyder application
    * :attr:`MonospaceInterface` is used, for instance, by the
      :guilabel:`Variable Explorer` and corresponds to the :attr:`Monospace`
      font resized to look good together with the :attr:`Interface` font.

    .. note::

        The values names in this enum are a result of historical reasons
        that date from Spyder 2 and are not easy to change now.
    """

    Monospace: str = "font"
    """Monospace font, used for code, output and ``literal/verbatim text``.

    Used in, for example, the :guilabel:`Editor`, :guilabel:`IPython Console`
    and :guilabel:`History` panes.
    """

    Interface: str = "app_font"
    """Interface font, used throughout the Spyder application."""

    MonospaceInterface: str = "monospace_app_font"
    """:attr:`Monospace` font resized to work with the :attr:`Interface` font.

    Used, for instance, by the :guilabel:`Variable Explorer` and corresponds
    to the :attr:`Monospace` font resized to look good when used together with
    the :attr:`Interface` font.
    """


class SpyderFontsMixin:
    """Mixin to get the different Spyder font types from our config system."""

    @classmethod
    def get_font(cls, font_type: str, font_size_delta: int = 0) -> QFont:
        """
        Get a font type as a :class:`QFont` object.

        Parameters
        ----------
        font_type: str
            A Spyder font type, one of the :class:`SpyderFontType` enum values.
        font_size_delta: int, optional
            Increase or decrease the default font size by this amount.
            The default is 0.
        """
        return get_font(option=font_type, font_size_delta=font_size_delta)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Toolbar Plugin.
"""

# Standard library imports
from typing import Union, Optional

# Third-party imports
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QWidget

# Local imports
from spyder.api.plugins import SpyderPluginV2, Plugins
from spyder.api.plugin_registration.decorators import (
    on_plugin_available, on_plugin_teardown)
from spyder.api.translations import _
from spyder.plugins.mainmenu.api import ApplicationMenus, WindowMenuSections
from spyder.plugins.toolbar.api import ApplicationToolbars
from spyder.plugins.toolbar.container import (
    ToolbarContainer, ToolbarMenus, ToolbarActions)
from spyder.utils.qthelpers import SpyderAction


class Toolbar(SpyderPluginV2):
    """
    Docstrings viewer widget.
    """
    NAME = 'toolbar'
    OPTIONAL = [Plugins.MainMenu]
    CONF_SECTION = NAME
    CONF_FILE = False
    CONTAINER_CLASS = ToolbarContainer
    CAN_BE_DISABLED = False

    # ---- SpyderPluginV2 API
    # -------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _('Toolbar')

    @staticmethod
    def get_description():
        return _('Manage application toolbars.')

    @classmethod
    def get_icon(cls):
        return QIcon()

    def on_initialize(self):
        create_app_toolbar = self.create_application_toolbar
        create_app_toolbar(ApplicationToolbars.File, _("File toolbar"))
        create_app_toolbar(ApplicationToolbars.Run, _("Run toolbar"))
        create_app_toolbar(ApplicationToolbars.Debug, _("Debug toolbar"))
        create_app_toolbar(ApplicationToolbars.Profile, _("Profile toolbar"))
        create_app_toolbar(ApplicationToolbars.Main, _("Main toolbar"))

    @on_plugin_available(plugin=Plugins.MainMenu)
    def on_main_menu_available(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)
        # Window menu Toolbar section
        mainmenu.add_item_to_application_menu(
            self.toolbars_menu,
            menu_id=ApplicationMenus.Window,
            section=WindowMenuSections.Toolbar,
            before_section=WindowMenuSections.Layout)
        mainmenu.add_item_to_application_menu(
            self.show_toolbars_action,
            menu_id=ApplicationMenus.Window,
            section=WindowMenuSections.Toolbar,
            before_section=WindowMenuSections.Layout)

    @on_plugin_teardown(plugin=Plugins.MainMenu)
    def on_main_menu_teardown(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)
        # Window menu Toolbar section
        mainmenu.remove_item_from_application_menu(
            ToolbarMenus.ToolbarsMenu,
            menu_id=ApplicationMenus.Window)
        mainmenu.remove_item_from_application_menu(
            ToolbarActions.ShowToolbars,
            menu_id=ApplicationMenus.Window)

    def on_mainwindow_visible(self):
        container = self.get_container()

        # Load all toolbars
        container.load_application_toolbars()

        # Create toolbars menu and show last visible toolbars
        container.create_toolbars_menu()
        container.load_last_visible_toolbars()

    def on_close(self, _unused):
        container = self.get_container()
        container.save_last_visible_toolbars()
        container.save_last_toolbars()
        for toolbar in container._visible_toolbars:
            toolbar.setVisible(False)

    # ---- Public API
    # ------------------------------------------------------------------------
    def create_application_toolbar(self, toolbar_id, title):
        """
        Create a Spyder application toolbar.

        Parameters
        ----------
        toolbar_id: str
            The toolbar unique identifier string.
        title: str
            The localized toolbar title to be displayed.

        Returns
        -------
        spyder.api.widgets.toolbar.ApplicationToolbar
            The created application toolbar.
        """
        toolbar = self.get_container().create_application_toolbar(
            toolbar_id, title)
        self.add_application_toolbar(toolbar)
        return toolbar

    def add_application_toolbar(self, toolbar):
        """
        Add toolbar to application toolbars.

        This can be used to add a custom toolbar. The `WorkingDirectory`
        plugin is an example of this.

        Parameters
        ----------
        toolbar: spyder.api.widgets.toolbars.ApplicationToolbar
            The application toolbar to add to the main window.
        """
        self.get_container().add_application_toolbar(toolbar, self._main)

    def remove_application_toolbar(self, toolbar_id: str):
        """
        Remove toolbar from the application toolbars.

        This can be used to remove a custom toolbar. The `WorkingDirectory`
        plugin is an example of this.

        Parameters
        ----------
        toolbar: str
            The application toolbar to remove from the main window.
        """
        self.get_container().remove_application_toolbar(toolbar_id, self._main)

    def add_item_to_application_toolbar(self,
                                        item: Union[SpyderAction, QWidget],
                                        toolbar_id: Optional[str] = None,
                                        section: Optional[str] = None,
                                        before: Optional[str] = None,
                                        before_section: Optional[str] = None,
                                        omit_id: bool = False):
        """
        Add action or widget `item` to given application menu `section`.

        Parameters
        ----------
        item: SpyderAction or QWidget
            The item to add to the `toolbar`.
        toolbar_id: str or None
            The application toolbar unique string identifier.
        section: str or None
            The section id in which to insert the `item` on the `toolbar`.
        before: str or None
            Make the item appear before another given item.
        before_section: str or None
            Make the item defined section appear before another given section
            (must be already defined).
        omit_id: bool
            If True, then the toolbar will check if the item to add declares an
            id, False otherwise. This flag exists only for items added on
            Spyder 4 plugins. Default: False
        """
        if before is not None:
            if not isinstance(before, str):
                raise ValueError('before argument must be a str')

        return self.get_container().add_item_to_application_toolbar(
                item,
                toolbar_id=toolbar_id,
                section=section,
                before=before,
                before_section=before_section,
                omit_id=omit_id
            )

    def remove_item_from_application_toolbar(self, item_id: str,
                                             toolbar_id: Optional[str] = None):
        """
        Remove action or widget `item` from given application menu by id.

        Parameters
        ----------
        item_id: str
            The item to remove from the toolbar.
        toolbar_id: str or None
            The application toolbar unique string identifier.
        """
        self.get_container().remove_item_from_application_toolbar(
            item_id,
            toolbar_id=toolbar_id
        )

    def get_application_toolbar(self, toolbar_id):
        """
        Return an application toolbar by toolbar_id.

        Parameters
        ----------
        toolbar_id: str
            The toolbar unique string identifier.

        Returns
        -------
        spyder.api.widgets.toolbars.ApplicationToolbar
            The application toolbar.
        """
        return self.get_container().get_application_toolbar(toolbar_id)

    def toggle_lock(self, value=None):
        """Lock/Unlock toolbars."""
        for toolbar in self.toolbarslist:
            toolbar.setMovable(not value)

    # ---- Convenience properties
    # -------------------------------------------------------------------------
    @property
    def toolbars_menu(self):
        return self.get_container().get_menu("toolbars_menu")

    @property
    def show_toolbars_action(self):
        return self.get_action("show toolbars")

    @property
    def toolbarslist(self):
        return self.get_container()._toolbarslist



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
spyder.plugins.remoteclient.api.modules
=======================================

Remote Client Plugin Modules API.

This package contains the API for the Remote Client Plugin
to interact with the Jupyter Server.

These modules provides an interface to control the Jupyter Server that sits
in the remote machine.
"""

from spyder.plugins.remoteclient.api.modules.base import JupyterAPI
from spyder.plugins.remoteclient.api.modules.file_services import (
    SpyderRemoteFileServicesAPI,
)

__all__ = ['JupyterAPI', 'SpyderRemoteFileServicesAPI']



# -----------------------------------------------------------------------------
# Copyright (c) 2021- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Spyder API plugin registration decorators.
"""

from __future__ import annotations

# Standard library imports
import functools
from collections.abc import Callable


def on_plugin_available(
    func: Callable | None = None,
    plugin: str | None = None,
) -> Callable:
    """
    Decorate a method to be called back when a specific plugin becomes ready.

    The decorated method must be a member of a
    :class:`~spyder.api.plugins.SpyderPluginV2` subclass for the decorator
    to work as intended.

    Methods to be decorated must have the signature

    .. code-block:: python

        def method(self):
            ...

    when observing a single plugin, or

    .. code-block:: python

        def method(self, plugin: str):
            ...

    when observing all plugins listed as dependencies.

    .. caution::

        Any ``plugin``\\(s) specified must be listed under either the
        :attr:`~spyder.api.plugins.SpyderPluginV2.REQUIRES` or
        :attr:`~spyder.api.plugins.SpyderPluginV2.OPTIONAL` class constants
        of the plugin's :class:`~spyder.api.plugins.SpyderPluginV2` class.
        If not, a :exc:`~spyder.api.exceptions.SpyderAPIError` will be raised
        when the plugin class is initialized.

    Parameters
    ----------
    func: Callable | None, optional
        Method to decorate, passed automatically when applying the decorator.
    plugin: str | None, optional
        Name of the requested plugin whose availability triggers ``func``.
        If ``None`` (the default), observes all plugins listed as dependencies.
        Must be listed under the class'
        ':attr:`~spyder.api.plugins.SpyderPluginV2.REQUIRES` or
        :attr:`~spyder.api.plugins.SpyderPluginV2.OPTIONAL` class constants,
        or else a :exc:`~spyder.api.exceptions.SpyderAPIError` will be raised.

    Returns
    -------
    func: Callable
        The method passed as ``func`` with the plugin listener set up.

    Raises
    ------
    SpyderAPIError
        When initializing a plugin class with decorated methods,
        if trying to watch a ``plugin`` that is not listed in the plugin class'
        :attr:`~spyder.api.plugins.SpyderPluginV2.REQUIRES` or
        :attr:`~spyder.api.plugins.SpyderPluginV2.OPTIONAL` class constants.
    """
    if func is None:
        return functools.partial(on_plugin_available, plugin=plugin)

    if plugin is None:
        # Use special __all identifier to signal that the function
        # observes all plugins listed as dependencies.
        plugin = "__all"

    func._plugin_listen = plugin
    return func


def on_plugin_teardown(
    func: Callable | None = None,
    plugin: str | None = None,
) -> Callable:
    """
    Decorate a method to be called back when tearing down a specific plugin.

    The decorated method must be a member of a
    :class:`~spyder.api.plugins.SpyderPluginV2` subclass for the decorator
    to work as intended.

    The decorator will be called **before** the specified ``plugin`` is deleted
    and also **before** the plugin that uses the decorator is destroyed.

    Methods that use this decorator must have the signature

    .. code-block:: python

        def method(self):
            ...

    .. important::

        A plugin name must be passed to ``plugin``. While a default of ``None``
        is accepted due to technical limitations, it will raise a
        :exc:`ValueError` at runtime.

    .. caution::

        Any ``plugin``\\(s) specified must be listed under either the
        :attr:`~spyder.api.plugins.SpyderPluginV2.REQUIRES` or
        :attr:`~spyder.api.plugins.SpyderPluginV2.OPTIONAL` class constants
        of the plugin's :class:`~spyder.api.plugins.SpyderPluginV2` class.
        If not, a :exc:`~spyder.api.exceptions.SpyderAPIError` will be raised
        when the plugin class is initialized.

    Parameters
    ----------
    func: Callable | None, optional
        Method to decorate, passed automatically when applying the decorator.
    plugin: str
        Name of the requested plugin whose teardown triggers ``func``.
        While ``None``, the default, is accepted for technical reasons,
        :exc:`ValueError` is raised if a plugin name is not passed.
        Must be listed under the class'
        ':attr:`~spyder.api.plugins.SpyderPluginV2.REQUIRES` or
        :attr:`~spyder.api.plugins.SpyderPluginV2.OPTIONAL` class constants,
        or else a :exc:`~spyder.api.exceptions.SpyderAPIError` will be raised.

    Returns
    -------
    func: Callable
        The method passed as ``func`` with the plugin listener set up.

    Raises
    ------
    ValueError
        If an explicit plugin name is not passed to ``plugin``.
    SpyderAPIError
        When initializing a plugin class with decorated methods,
        if trying to watch a ``plugin`` that is not listed in the plugin class'
        :attr:`~spyder.api.plugins.SpyderPluginV2.REQUIRES` or
        :attr:`~spyder.api.plugins.SpyderPluginV2.OPTIONAL` class constants.
    """
    if func is None:
        return functools.partial(on_plugin_teardown, plugin=plugin)

    if plugin is None:
        raise ValueError(
            "on_plugin_teardown must have a well defined "
            "plugin keyword argument value, "
            "e.g., plugin=Plugins.Editor"
        )

    func._plugin_teardown = plugin
    return func



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------
"""
Spyder WebSocket Kernel Client.

This module implements a WebSocket client for Spyder kernels.
"""

from __future__ import annotations

import asyncio
import logging
import os
import typing as t
from functools import wraps
from getpass import getuser
from types import MethodType

import aiohttp
from aiohttp.client_exceptions import ClientConnectionResetError
from jupyter_client.adapter import adapt
from jupyter_client.channels import major_protocol_version
from jupyter_client.client import validate_string_dict
from jupyter_client.jsonutil import extract_dates
from jupyter_client.session import (
    extract_header,
    json_packer,
    json_unpacker,
    msg_header,
    new_id,
)
from qtconsole.kernel_mixins import QtKernelClientMixin
from qtconsole.util import SuperQObject
from qtpy.QtCore import Signal
from traitlets.config import Configurable

from spyder.api.asyncdispatcher import AsyncDispatcher

_LOGGER = logging.getLogger(__name__)


class _Session:
    def __init__(
        self,
        session: t.Optional[str] = None,
        username: t.Optional[str] = None,
        adapt_version: int = major_protocol_version,
        metadata: t.Optional[dict[str, t.Any]] = None,
        check_pid: bool = True,
    ):
        self.session = session or new_id()
        self.username = username or getuser()
        self.check_pid = check_pid
        self.adapt_version = adapt_version
        self.pid = os.getpid()
        self.none = self.pack({})
        self.message_count = 0
        self.metadata = metadata or {}

    @property
    def bsession(self):
        return self.session.encode("ascii")

    @staticmethod
    def pack(msg: dict[str, t.Any]) -> bytes:
        return json_packer(msg)

    @staticmethod
    def unpack(msg: bytes) -> dict[str, t.Any]:
        return json_unpacker(msg)

    async def send(
        self,
        stream: aiohttp.ClientWebSocketResponse,
        channel: str,
        msg: dict[str, t.Any],
    ) -> dict[str, t.Any] | None:
        """
        Build and send a message via websocket.

        The message format used by this function internally is as follows:

        [p_header, p_parent, p_content, buffer1, buffer2,...]

        The serialize/deserialize methods convert the nested message dict into
        this format.

        Parameters
        ----------
        stream: aiohttp.ClientWebSocketResponse
            The websocket stream to send the message to.

        msg : Message/dict
            The message to send. This should be a dict with the keys
            [header, parent_header, metadata, content, buffers].

        Returns
        -------
        msg : dict
            The constructed message.

        Raises
        ------
        TypeError
            If the buffers do not support the buffer protocol.
        ValueError
            If the buffers are not contiguous.
        """
        buffers = msg.get("buffers", [])

        if self.check_pid and os.getpid() != self.pid:
            _LOGGER.warning(
                "WARNING: attempted to send message from fork\n%s", msg
            )
            return None

        buffers = [] if buffers is None else buffers
        for idx, buf in enumerate(buffers):
            if isinstance(buf, memoryview):
                view = buf
            else:
                try:
                    # check to see if buf supports the buffer protocol.
                    view = memoryview(buf)
                except TypeError as e:
                    emsg = "Buffer objects must support the buffer protocol."
                    raise TypeError(emsg) from e

            # memoryview.contiguous is new in 3.3, just skip the check on
            # Python 2
            if hasattr(view, "contiguous") and not view.contiguous:
                # zmq requires memoryviews to be contiguous
                msge = f"Buffer {idx} ({buf}) is not contiguous"
                raise ValueError(msge)

        if self.adapt_version:
            msg = adapt(msg, self.adapt_version)

        to_send = self.serialize(msg)
        to_send.extend(buffers)

        try:
            await stream.send_bytes(
                self._serialize_components_v1_protocol(to_send, channel),
            )
        except ClientConnectionResetError:
            # Connection was closed so it's not possible to send msg
            return {}

        return msg

    async def recv(
        self,
        stream: aiohttp.ClientWebSocketResponse,
        timeout: float | None = None,
    ) -> tuple[str, dict[str, t.Any] | dict[int, t.Any]]:
        """
        Receive a message from the websocket stream.

        The message format used by this function internally is as follows:

        [p_header, p_parent, p_content, buffer1, buffer2,...]

        The serialize/deserialize methods convert the nested message dict into
        this format.

        Parameters
        ----------
        stream: aiohttp.ClientWebSocketResponse
            The websocket stream to receive the message from.

        Returns
        -------
        channel: str
            The channel name of the message.
        msg: dict
            The message dict.

        Raises
        ------
        aiohttp.WSMessageTypeError
            If the received message is not of type WSMsgType.BINARY.
        """
        msg = await stream.receive(timeout=timeout)
        if msg.type in {aiohttp.WSMsgType.PING, aiohttp.WSMsgType.PONG}:
            return "hb", {msg.type: msg.data}

        if msg.type in {aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.CLOSING}:
            return "closed", {}

        if msg.type is not aiohttp.WSMsgType.BINARY:
            msg = (
                f"Received message {msg.type}:{msg.data!r}"
                f" is not WSMsgType.BINARY"
            )
            raise aiohttp.WSMessageTypeError(msg)

        channel, components = self._deserialize_components_v1_protocol(msg.data)
        return channel, self.deserialize(components)

    def serialize(
        self,
        msg: dict[str, t.Any],
    ) -> list[bytes]:
        """
        Serialize the message components to bytes.

        This is roughly the inverse of deserialize. The serialize/deserialize
        methods work with full message lists, whereas pack/unpack work with
        the individual message parts in the message list.

        Parameters
        ----------
        msg: dict or Message
            The next message dict as returned by the self.msg method.

        Returns
        -------
        msg_list : list
            The list of bytes objects to be sent with the format::

                [p_header, p_parent, p_metadata,
                 p_content, buffer1, buffer2, ...]

            In this list, the ``p_*`` entities are the packed or serialized
            versions, so if JSON is used, these are utf8 encoded JSON strings.

        Raises
        ------
        TypeError
            If the message is malformed.
        """
        content = msg.get("content", {})
        if content is None:
            content = self.none
        elif isinstance(content, dict):
            content = self.pack(content)
        elif isinstance(content, bytes):
            # content is already packed, as in a relayed message
            pass
        elif isinstance(content, str):
            # should be bytes, but JSON often spits out unicode
            content = content.encode("utf8")
        else:
            emsg = f"Content incorrect type: {type(content)}"
            raise TypeError(emsg)

        return [
            self.pack(msg["header"]),
            self.pack(msg["parent_header"]),
            self.pack(msg["metadata"]),
            content,
        ]

    def deserialize(
        self,
        msg_list: list[bytes],
        content: bool = True,
    ) -> dict[str, t.Any]:
        """
        Unserialize a msg_list to a nested message dict.

        This is roughly the inverse of serialize. The serialize/deserialize
        methods work with full message lists, whereas pack/unpack work with
        the individual message parts in the message list.

        Parameters
        ----------
        msg_list : list of bytes or Message objects
            The list of message parts of the form [p_header,p_parent,
            p_metadata,p_content,buffer1,buffer2,...].
        content : bool (True)
            Whether to unpack the content dict (True), or leave it packed
            (False).

        Returns
        -------
        msg : dict
            The nested message dict with top-level keys [header, parent_header,
            content, buffers].  The buffers are returned as memoryviews.

        Raises
        ------
        TypeError
            If the message is malformed.
        """
        minlen = 4
        message = {}
        if not len(msg_list) >= minlen:
            msg = f"malformed message, must have at least {minlen} elements"
            raise TypeError(msg)

        header = self.unpack(msg_list[0])
        message["header"] = extract_dates(header)
        message["msg_id"] = header["msg_id"]
        message["msg_type"] = header["msg_type"]
        message["parent_header"] = extract_dates(self.unpack(msg_list[1]))
        message["metadata"] = self.unpack(msg_list[2])

        if content:
            message["content"] = self.unpack(msg_list[3])
        else:
            message["content"] = msg_list[3]

        buffers = [memoryview(b) for b in msg_list[4:]]
        message["buffers"] = buffers

        return adapt(message)

    @property
    def msg_id(self) -> str:
        message_number = self.message_count
        self.message_count += 1
        return f"{self.session}_{os.getpid()}_{message_number}"

    def msg_header(self, msg_type: str) -> dict[str, t.Any]:
        """Create a header for a message type."""
        return msg_header(self.msg_id, msg_type, self.username, self.session)

    def msg(
        self,
        msg_type: str,
        content: dict | None = None,
        parent: dict[str, t.Any] | None = None,
        header: dict[str, t.Any] | None = None,
        metadata: dict[str, t.Any] | None = None,
    ) -> dict[str, t.Any]:
        """
        Return the nested message dict.

        This format is different from what is sent over the wire. The
        serialize/deserialize methods converts this nested message dict to the
        wire format, which is a list of message parts.
        """
        msg = {}
        header = self.msg_header(msg_type) if header is None else header
        msg["header"] = header
        msg["msg_id"] = header["msg_id"]
        msg["msg_type"] = header["msg_type"]
        msg["parent_header"] = {} if parent is None else extract_header(parent)
        msg["content"] = {} if content is None else content
        msg["metadata"] = self.metadata.copy()

        if metadata is not None:
            msg["metadata"].update(metadata)

        return msg

    @staticmethod
    def _serialize_components_v1_protocol(
        components: list[bytes], channel: str
    ) -> bytes:
        echannel = channel.encode("utf-8")
        offsets = [8 * (1 + 1 + len(components) + 1)]
        offsets.append(len(echannel) + offsets[-1])

        for msg in components:
            offsets.append(len(msg) + offsets[-1])

        offset_number = len(offsets).to_bytes(8, byteorder="little")
        offsets = [
            offset.to_bytes(8, byteorder="little") for offset in offsets
        ]

        return b"".join([offset_number, *offsets, echannel, *components])

    @staticmethod
    def _deserialize_components_v1_protocol(
        ws_msg: bytes,
    ) -> tuple[str, list[bytes]]:
        offset_number = int.from_bytes(ws_msg[:8], "little")
        offsets = [
            int.from_bytes(ws_msg[8 * (i + 1) : 8 * (i + 2)], "little")
            for i in range(offset_number)
        ]
        channel = ws_msg[offsets[0] : offsets[1]].decode("utf-8")
        msg_list = [
            ws_msg[offsets[i] : offsets[i + 1]]
            for i in range(1, offset_number - 1)
        ]

        return channel, msg_list

    def send_raw(self, *args, **kwargs):
        msg = "send_raw is not implemented for WebSocket connections"
        raise NotImplementedError(msg)

    def feed_identities(self, *args, **kwargs):
        msg = "Websocket protocol does not support signing"
        raise NotImplementedError(msg)

    def sign(self, *args, **kwargs):
        msg = "Websocket protocol does not support signing"
        raise NotImplementedError(msg)


class _ChannelQueues:
    def __init__(self):
        AsyncDispatcher(loop="ipythonconsole", early_return=False)(
            self._create_queue
        )()

    async def _create_queue(self):
        self.shell = asyncio.Queue()
        self.iopub = asyncio.Queue()
        self.stdin = asyncio.Queue()
        self.control = asyncio.Queue()
        self.hb = asyncio.Queue()

    def __getitem__(self, channel: str) -> asyncio.Queue[dict]:
        return getattr(self, channel)


class _WebSocketChannel:
    def __init__(
        self,
        queue: asyncio.Queue[dict],
        websocket: aiohttp.ClientWebSocketResponse,
        session: _Session,
        channel_name: str,
    ):
        self._queue = queue
        self._websocket = websocket
        self.session = session
        self.channel_name = channel_name
        self._running: asyncio.Task[t.NoReturn] | None = None
        self._inspect = None

    async def _loop(self):
        """Receive messages from the websocket stream."""
        while True:
            msg = await self._queue.get()
            self.call_handlers(msg)
            if self._inspect is not None:
                self._inspect(msg)
            self._queue.task_done()

    def start(self):
        """Start the channel."""
        _LOGGER.debug("Starting channel %s", self.channel_name)
        self._running = asyncio.create_task(
            self._loop(), name=f"{self.session.session}-{self.channel_name}"
        )

    async def stop(self) -> None:
        """Stop the channel."""
        if self._running is not None:
            await self._queue.join()
            self._running.cancel()
            self._running = None

    def is_alive(self) -> bool:
        """Test whether the channel is alive."""
        return self._running is not None and not self._running.done()

    @AsyncDispatcher(loop="ipythonconsole", early_return=False)
    async def send(self, msg: dict[str, t.Any]) -> None:
        """Send a message to the channel."""
        await self.session.send(self._websocket, self.channel_name, msg)

    def call_handlers(self, msg: dict):
        pass


class _FakeHBChannel:
    time_to_dead: float = 0

    def __init__(self, *args, **kwargs):
        self._beating = True

    def pause(self):
        """Pause the heartbeat channel."""
        pass

    def unpause(self):
        """Unpause the heartbeat channel."""
        pass

    def is_beating(self) -> bool:
        """Test whether the channel is beating."""
        return self._beating

    def call_handlers(self, since_last_heartbeat: float):
        pass


class _WebSocketKernelClient(Configurable):
    """Pure-async client for remote WS Jupyter kernels."""

    shell_channel_class = _WebSocketChannel  # type: ignore[assignment]
    iopub_channel_class = _WebSocketChannel  # type: ignore[assignment]
    hb_channel_class = _FakeHBChannel  # type: ignore[assignment]
    stdin_channel_class = _WebSocketChannel  # type: ignore[assignment]
    control_channel_class = _WebSocketChannel  # type: ignore[assignment]

    def __init__(
        self,
        endpoint: str,
        token: str | None = None,
        username: str | None = None,
        session_id: str | None = None,
        aiohttp_session: aiohttp.ClientSession | None = None,
    ):
        self._endpoint = endpoint
        self._token = token

        self.session = _Session(
            session=session_id,
            username=username,
        )

        self._aiohttp_session = aiohttp_session
        self._owns_session = aiohttp_session is None
        self._ws: aiohttp.ClientWebSocketResponse | None = None
        self._receiver: asyncio.Task[None] | None = None

        self._queues = _ChannelQueues()

        self._shell_channel: _WebSocketChannel | None = None
        self._iopub_channel: _WebSocketChannel | None = None
        self._stdin_channel: _WebSocketChannel | None = None
        self._control_channel: _WebSocketChannel | None = None
        self._hb_channel: _FakeHBChannel | None = None

    @property
    def allow_stdin(self) -> bool:
        """Whether the kernel allows stdin requests."""
        return self.stdin_channel.is_alive()

    @property
    def shell_channel(self):
        if self._ws is None:
            msg = "WebSocket connection is not established."
            raise RuntimeError(msg)

        if self._shell_channel is None:
            self._shell_channel = self.shell_channel_class(
                self._queues.shell,
                self._ws,
                self.session,
                "shell",
            )
        return self._shell_channel

    @property
    def iopub_channel(self):
        if self._ws is None:
            msg = "WebSocket connection is not established."
            raise RuntimeError(msg)

        if self._iopub_channel is None:
            self._iopub_channel = self.iopub_channel_class(
                self._queues.iopub,
                self._ws,
                self.session,
                "iopub",
            )
        return self._iopub_channel

    @property
    def stdin_channel(self):
        if self._ws is None:
            msg = "WebSocket connection is not established."
            raise RuntimeError(msg)

        if self._stdin_channel is None:
            self._stdin_channel = self.stdin_channel_class(
                self._queues.stdin,
                self._ws,
                self.session,
                "stdin",
            )
        return self._stdin_channel

    @property
    def control_channel(self):
        if self._ws is None:
            msg = "WebSocket connection is not established."
            raise RuntimeError(msg)

        if self._control_channel is None:
            self._control_channel = self.control_channel_class(
                self._queues.control,
                self._ws,
                self.session,
                "control",
            )
        return self._control_channel

    @property
    def hb_channel(self):
        if self._ws is None:
            msg = "WebSocket connection is not established."
            raise RuntimeError(msg)

        if self._hb_channel is None:
            self._hb_channel = self.hb_channel_class()
        return self._hb_channel

    @property
    def kernel(self):
        return None  # remote kernel (no local process handle)

    @property
    def channels_running(self) -> bool:
        """Check if any of the channels created and running."""
        return (
            (bool(self._shell_channel) and self._shell_channel.is_alive())
            or (bool(self._iopub_channel) and self._iopub_channel.is_alive())
            or (bool(self._stdin_channel) and self._stdin_channel.is_alive())
            or (
                bool(self._control_channel)
                and self._control_channel.is_alive()
            )
        )

    @AsyncDispatcher(loop="ipythonconsole", early_return=False)
    async def start_channels(
        self,
        shell: bool = True,
        iopub: bool = True,
        stdin: bool = True,
        hb: bool = True,
        control: bool = True,
    ):
        """
        Start the channels.

        Parameters
        ----------
        shell : bool, optional
            Whether to start the shell channel. Default is True.
        iopub : bool, optional
            Whether to start the iopub channel. Default is True.
        stdin : bool, optional
            Whether to start the stdin channel. Default is True.
        hb : bool, optional
            Whether to start the heartbeat channel. Default is True.
        control : bool, optional
            Whether to start the control channel. Default is True.
        """
        _LOGGER.info("Starting channels for %s", self.session.session)
        await self._connect()

        if shell:
            self.shell_channel.start()
            self.shell_channel._websocket = self._ws
            self._shell_channel._inspect = self._check_kernel_info_reply
        if iopub:
            self.iopub_channel.start()
            self.iopub_channel._websocket = self._ws
        if stdin:
            self.stdin_channel.start()
            self.stdin_channel._websocket = self._ws
        if control:
            self.control_channel.start()
            self.control_channel._websocket = self._ws

        self._receiver = asyncio.create_task(
            self._receiver_loop(), name=f"{self.session.session}-ws-receiver"
        )

        _LOGGER.debug("Websocket client started for %s", self.session.session)

    def _check_kernel_info_reply(self, msg: dict[str, t.Any]):
        if msg["msg_type"] == "kernel_info_reply":
            self._handle_kernel_info_reply(msg)
            self._shell_channel._inspect = None

    async def _connect(self):
        if self._aiohttp_session is None:
            self._owns_session = True
            self._aiohttp_session = aiohttp.ClientSession()

        qs = {"session_id": self.session.session}
        if self._token:
            qs["token"] = self._token

        _LOGGER.info("Connecting to Websocket at: %s", self._endpoint)
        self._ws = await self._aiohttp_session.ws_connect(
            self._endpoint,
            params=qs,
            protocols=("v1.kernel.websocket.jupyter.org",),
            autoping=True,
            autoclose=True,
            max_msg_size=104857600,  # 100 MB
        )

    async def _receiver_loop(self):
        """Receive messages from the websocket stream."""
        if self._ws is None:
            msg = "WebSocket connection is not established."
            raise RuntimeError(msg)
        try:
            while True:
                channel, msg = await self.session.recv(self._ws)
                if channel == "closed":
                    break
                await self._queues[channel].put(msg)
        except asyncio.CancelledError:
            _LOGGER.debug(
                "Receiver loop cancelled for %s", self.session.session
            )
        except Exception as exc:
            await self._ws.close(code=aiohttp.WSCloseCode.INTERNAL_ERROR)
            self._handle_receiver_exception(exc)

    @staticmethod
    def _handle_receiver_exception(exc: Exception):
        """Handle exceptions in the receiver loop."""
        raise exc

    @AsyncDispatcher(loop="ipythonconsole", early_return=False)
    async def stop_channels(self):
        """Stop the channels."""
        _LOGGER.info("Stopping channels for %s", self.session.session)

        if self._receiver is not None:
            self._receiver.cancel()
            self._receiver = None

        if self.shell_channel.is_alive():
            await self.shell_channel.stop()
        if self.iopub_channel.is_alive():
            await self.iopub_channel.stop()
        if self.stdin_channel.is_alive():
            await self.stdin_channel.stop()
        if self.control_channel.is_alive():
            await self.control_channel.stop()

        await self._disconnect()

        _LOGGER.debug("Disconnected from %s", self.session.session)

    async def _disconnect(self):
        if self._ws is not None:
            await self._ws.close()

        if self._owns_session and self._aiohttp_session is not None:
            await self._aiohttp_session.close()

    def _handle_kernel_info_reply(self, msg: dict[str, t.Any]) -> None:
        """
        Handle kernel info reply.

        It sets protocol adaptation version. This might be run from a separate
        thread.
        """
        adapt_version = int(msg["content"]["protocol_version"].split(".")[0])
        if adapt_version != major_protocol_version:
            self.session.adapt_version = adapt_version

    def is_alive(self) -> bool:
        """Test whether the kernel is alive."""
        return self._ws is not None and not self._ws.closed

    def execute(
        self,
        code: str,
        silent: bool = False,
        store_history: bool = True,
        user_expressions: dict[str, t.Any] | None = None,
        allow_stdin: bool | None = None,
        stop_on_error: bool = True,
    ) -> str:
        """
        Execute code in the kernel.

        Parameters
        ----------
        code : str
            A string of code in the kernel's language.
        silent : bool, optional (default False)
            If set, the kernel will execute the code as quietly possible, and
            will force store_history to be False.
        store_history : bool, optional (default True)
            If set, the kernel will store command history. This is forced
            to be False if silent is True.
        user_expressions : dict, optional
            A dict mapping names to expressions to be evaluated in the user's
            dict. The expression values are returned as strings formatted using
            :func:`repr`.
        allow_stdin : bool, optional (default self.allow_stdin)
            Flag for whether the kernel can send stdin requests to frontends.
            Some frontends (e.g. the Notebook) do not support stdin requests.
            If raw_input is called from code executed from such a frontend, a
            StdinNotImplementedError will be raised.
        stop_on_error: bool, optional (default True)
            Flag whether to abort the execution queue, if an exception is
            encountered.

        Returns
        -------
        The msg_id of the message sent.

        Raises
        ------
        TypeError
            If the code is not a string.
        """
        if user_expressions is None:
            user_expressions = {}
        if allow_stdin is None:
            allow_stdin = self.allow_stdin

        # Don't waste network traffic if inputs are invalid
        if not isinstance(code, str):
            msg = f"code {code!r} must be a string"
            raise TypeError(msg)
        validate_string_dict(user_expressions)

        # Create class for content/msg creation. Related to, but possibly
        # not in Session.
        content = {
            "code": code,
            "silent": silent,
            "store_history": store_history,
            "user_expressions": user_expressions,
            "allow_stdin": allow_stdin,
            "stop_on_error": stop_on_error,
        }
        msg = self.session.msg("execute_request", content)
        self.shell_channel.send(msg)
        return msg["header"]["msg_id"]

    def complete(self, code: str, cursor_pos: t.Optional[int] = None) -> str:
        """
        Tab complete text in the kernel's namespace.

        Parameters
        ----------
        code : str
            The context in which completion is requested. It can be anything
            between a variable name and an entire cell.
        cursor_pos : int, optional
            The position of the cursor in the block of code where the
            completion was requested. The default is ``len(code)``.

        Returns
        -------
        The msg_id of the message sent.
        """
        if cursor_pos is None:
            cursor_pos = len(code)
        content = {"code": code, "cursor_pos": cursor_pos}
        msg = self.session.msg("complete_request", content)
        self.shell_channel.send(msg)
        return msg["header"]["msg_id"]

    def inspect(
        self,
        code: str,
        cursor_pos: int | None = None,
        detail_level: int = 0,
    ) -> str:
        """
        Get metadata information about an object in the kernel's namespace.

        It is up to the kernel to determine the appropriate object to inspect.

        Parameters
        ----------
        code : str
            The context in which info is requested. It can be anything between
            a variable name and an entire cell.
        cursor_pos : int, optional
            The position of the cursor in the block of code where the info was
            requested. The default is ``len(code)``.
        detail_level : int, optional
            The level of detail for the introspection (0-2).

        Returns
        -------
        The msg_id of the message sent.
        """
        if cursor_pos is None:
            cursor_pos = len(code)
        content = {
            "code": code,
            "cursor_pos": cursor_pos,
            "detail_level": detail_level,
        }
        msg = self.session.msg("inspect_request", content)
        self.shell_channel.send(msg)
        return msg["header"]["msg_id"]

    def history(
        self,
        raw: bool = True,
        output: bool = False,
        hist_access_type: str = "range",
        **kwargs,
    ) -> str:
        """
        Get entries from the kernel's history list.

        Parameters
        ----------
        raw : bool
            If True, return the raw input.
        output : bool
            If True, then return the output as well.
        hist_access_type : str
            'range' (fill in session, start and stop params), 'tail' (fill in
             n) or 'search' (fill in pattern param).

        session : int
            For a range request, the session from which to get lines. Session
            numbers are positive integers; negative ones count back from the
            current session.
        start : int
            The first line number of a history range.
        stop : int
            The final (excluded) line number of a history range.
        n : int
            The number of lines of history to get for a tail request.
        pattern : str
            The glob-syntax pattern for a search request.

        Returns
        -------
        The ID of the message sent.
        """
        if hist_access_type == "range":
            kwargs.setdefault("session", 0)
            kwargs.setdefault("start", 0)
        content = dict(
            raw=raw, output=output, hist_access_type=hist_access_type, **kwargs
        )
        msg = self.session.msg("history_request", content)
        self.shell_channel.send(msg)
        return msg["header"]["msg_id"]

    def kernel_info(self) -> str:
        """
        Request kernel info.

        Returns
        -------
        The msg_id of the message sent.
        """
        msg = self.session.msg("kernel_info_request")
        self.shell_channel.send(msg)
        return msg["header"]["msg_id"]

    def comm_info(self, target_name: str | None = None) -> str:
        """
        Request comm info.

        Returns
        -------
        The msg_id of the message sent.
        """
        content = {} if target_name is None else {"target_name": target_name}
        msg = self.session.msg("comm_info_request", content)
        self.shell_channel.send(msg)
        return msg["header"]["msg_id"]

    def is_complete(self, code: str) -> str:
        """
        Ask the kernel whether some code is complete and ready to execute.

        Returns
        -------
        The ID of the message sent.
        """
        msg = self.session.msg("is_complete_request", {"code": code})
        self.shell_channel.send(msg)
        return msg["header"]["msg_id"]

    def input(self, string: str) -> None:
        """
        Send a string of raw input to the kernel.

        This should only be called in response to the kernel sending an
        ``input_request`` message on the stdin channel.

        Returns
        -------
        The ID of the message sent.
        """
        content = {"value": string}
        msg = self.session.msg("input_reply", content)
        self.stdin_channel.send(msg)
        return msg["header"]["msg_id"]

    def shutdown(self, restart: bool = False) -> str:
        """
        Request an immediate kernel shutdown on the control channel.

        Upon receipt of the (empty) reply, client code can safely assume that
        the kernel has shut down and it's safe to forcefully terminate it if
        it's still alive.

        The kernel will send the reply via a function registered with Python's
        atexit module, ensuring it's truly done as the kernel is done with all
        normal operation.

        Returns
        -------
        The msg_id of the message sent
        """
        # Send quit message to kernel. Once we implement kernel-side setattr,
        # this should probably be done that way, but for now this will do.
        msg = self.session.msg("shutdown_request", {"restart": restart})
        self.control_channel.send(msg)
        return msg["header"]["msg_id"]


class QtWSHBChannel(_FakeHBChannel, SuperQObject):
    """A heartbeat channel emitting a Qt signal when a message is received."""

    # Never used, but kept for API compatibility
    kernel_died = Signal(float)


class QtWSChannel(_WebSocketChannel, SuperQObject):
    """A channel emitting a Qt signal when a message is received."""

    message_received = Signal(object)

    def call_handlers(self, msg):
        """
        Emit signal when a message is received.

        This method is called in the ioloop thread when a message arrives.

        It is important to remember that this method is called in the thread
        so that some logic must be done to ensure that the application level
        handlers are called in the application thread.
        """
        self.message_received.emit(msg)

    def closed(self):
        """Check if the channel is closed."""
        return not self.is_alive()

    def flush(self):
        pass


class SpyderWSKernelClient(QtKernelClientMixin, _WebSocketKernelClient):
    """A KernelClient that provides signals and slots."""

    iopub_channel_class = QtWSChannel
    shell_channel_class = QtWSChannel
    stdin_channel_class = QtWSChannel
    hb_channel_class = QtWSHBChannel
    control_channel_class = QtWSChannel

    sig_spyder_kernel_info = Signal(object)

    sig_ws_msg_size_overflow = Signal(int)

    def _handle_kernel_info_reply(self, rep):
        """Check spyder-kernels version."""
        super()._handle_kernel_info_reply(rep)
        spyder_kernels_info = rep["content"].get("spyder_kernels_info", None)
        self.sig_spyder_kernel_info.emit(spyder_kernels_info)

    def _handle_receiver_exception(self, exc: BaseException):
        """Handle exceptions in the receiver loop."""
        if (
            isinstance(exc, aiohttp.WebSocketError)
            and exc.code == aiohttp.WSCloseCode.MESSAGE_TOO_BIG
        ):
            _LOGGER.error(
                "WebSocket message too big for %s: %s",
                self.session.session,
                exc,
            )
            self.sig_ws_msg_size_overflow.emit(self._ws._reader._max_msg_size)
        else:
            _LOGGER.error(
                "Receiver loop exception for %s: %s",
                self.session.session,
                exc,
            )
            super()._handle_receiver_exception(exc)



# -----------------------------------------------------------------------------
# Copyright (c) 2021- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Mixins for plugins/widgets that are connected to the IPython console.
"""

from __future__ import annotations

# Standard library imports
from typing import TYPE_CHECKING

# Third party imports
from qtpy.QtCore import Signal

# Local imports
from spyder.api.plugin_registration.decorators import (
    on_plugin_available,
    on_plugin_teardown,
)
from spyder.api.plugins import Plugins

if TYPE_CHECKING:
    from qtpy.QtWidgets import QWidget

    import spyder.plugins.ipythonconsole.plugin
    import spyder.plugins.ipythonconsole.widgets


class ShellConnectMixin:
    """
    Mixin to connect any widget or object to an IPython Console shell widget.
    """

    # ---- Connection to the IPython console
    # -------------------------------------------------------------------------
    def register_ipythonconsole(
        self,
        ipythonconsole: spyder.plugins.ipythonconsole.plugin.IPythonConsole,
    ) -> None:
        """
        Connect this object to the relevant signals of the IPython Console.

        Parameters
        ----------
        ipythonconsole : spyder.plugins.ipythonconsole.plugin.IPythonConsole
            The IPython Console to connect this object to the signals of.

        Returns
        -------
        None
        """
        ipythonconsole.sig_shellwidget_changed.connect(self.set_shellwidget)
        ipythonconsole.sig_shellwidget_created.connect(self.add_shellwidget)
        ipythonconsole.sig_shellwidget_deleted.connect(self.remove_shellwidget)
        ipythonconsole.sig_shellwidget_errored.connect(
            self.add_errored_shellwidget
        )

    def unregister_ipythonconsole(
        self,
        ipythonconsole: spyder.plugins.ipythonconsole.plugin.IPythonConsole,
    ) -> None:
        """
        Disconnect this object from the relevant signals of the IPython Console.

        Parameters
        ----------
        ipythonconsole : spyder.plugins.ipythonconsole.plugin.IPythonConsole
            The IPython Console to disconnect this object from the signals of.

        Returns
        -------
        None
        """
        ipythonconsole.sig_shellwidget_changed.disconnect(self.set_shellwidget)
        ipythonconsole.sig_shellwidget_created.disconnect(self.add_shellwidget)
        ipythonconsole.sig_shellwidget_deleted.disconnect(
            self.remove_shellwidget
        )
        ipythonconsole.sig_shellwidget_errored.disconnect(
            self.add_errored_shellwidget
        )

    # ---- Public API
    # -------------------------------------------------------------------------
    def set_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """
        Set as active the stack widget associated to the given shell widget.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget that corresponds to the stacked widget to set
            as currently active.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def add_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """
        Create a new object associated with the given shell widget.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget to associate the new widget to.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def remove_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """
        Remove the object associated to a given shell widget.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget to remove the associated object of.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def add_errored_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """
        Register an error widget if the shell widget's kernel failed to start.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget to associate with a new error widget.

        Returns
        -------
        None
        """
        raise NotImplementedError


class ShellConnectWidgetForStackMixin:
    """
    Mixin for widgets added to a stack of those connected to different shells.

    Used for the stacked widgets of
    :class:`~spyder.api.shellconnect.main_widget.ShellConnectMainWidget`.
    """

    sig_show_empty_message_requested: Signal = Signal(bool)
    """
    Request that the empty message be shown/hidden.

    Parameters
    ----------
    show_empty_message: bool
        ``True`` to show the empty message; ``False`` for the normal widget.
    """

    def __init__(self) -> None:
        """
        Set up attributes on the mixin.

        Returns
        -------
        None
        """

        self.is_empty: bool = True
        """
        If the current widget has content to show.

        This attribute is necessary to track whether this widget has content
        to display.
        """


class ShellConnectPluginMixin(ShellConnectMixin):
    """
    Mixin to connect a plugin composed of stacked widgets to the shell widgets
    in the IPython console.

    It is assumed that self.get_widget() returns an instance of
    ShellConnectMainWidget.
    """

    # ---- Connection to the IPython console
    # -------------------------------------------------------------------------
    @on_plugin_available(plugin=Plugins.IPythonConsole)
    def on_ipython_console_available(self) -> None:
        """
        Connect this object to the relevant signals of the IPython Console.

        Returns
        -------
        None
        """
        ipythonconsole = self.get_plugin(Plugins.IPythonConsole)
        self.register_ipythonconsole(ipythonconsole)

    @on_plugin_teardown(plugin=Plugins.IPythonConsole)
    def on_ipython_console_teardown(self) -> None:
        """
        Disconnect this object from the relevant signals of the IPython Console.

        Returns
        -------
        None
        """
        ipythonconsole = self.get_plugin(Plugins.IPythonConsole)
        self.unregister_ipythonconsole(ipythonconsole)

    # ---- Public API
    # -------------------------------------------------------------------------
    def set_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """
        Set as active the stack widget associated to the given shell widget.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget that corresponds to the stacked widget to set
            as currently active.

        Returns
        -------
        None
        """
        self.get_widget().set_shellwidget(shellwidget)

    def add_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """
        Create a new widget in the stack associated to a given shell widget.

        This method registers a new widget to display the content that
        is associated with the given shell widget.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget to associate the new widget to.

        Returns
        -------
        None
        """
        self.get_widget().add_shellwidget(shellwidget)

    def remove_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """
        Remove the widget associated to a given shell widget.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget to remove the associated widget of.

        Returns
        -------
        None
        """
        self.get_widget().remove_shellwidget(shellwidget)

    def add_errored_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """
        Add an error widget for a shell widget whose kernel failed to start.

        This is necessary to show a meaningful message when switching to
        consoles with dead kernels.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget to associate with a new error widget.

        Returns
        -------
        None
        """
        self.get_widget().add_errored_shellwidget(shellwidget)

    def current_widget(self) -> QWidget:
        """
        Get the widget corresponding to the currently active console tab.

        Returns
        -------
        QWidget
            The current widget in the stack, associated with the active
            :class:`~spyder.plugins.ipythonconsole.widgets.ShellWidget`
            (i.e. :guilabel:`IPython Console` tab).
        """
        return self.get_widget().current_widget()

    def get_widget_for_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> QWidget:
        """
        Retrieve the stacked widget corresponding to the given shell widget.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget to return the associated widget of.

        Returns
        -------
        QWidget | None
            The widget in the stack associated with ``shellwidget``,
            or ``None`` if not found.
        """
        return self.get_widget().get_widget_for_shellwidget(shellwidget)
